import os
import pandas as pd
from astroquery.vizier import Vizier

from ..utils import local_package_path

def clear_downloaded_files() -> None:
    """Delete all files in local folder."""
    for file in os.listdir(local_package_path('downloaded_catalogues')):
        if file.endswith('.pkl'):
            os.remove(local_package_path(f'downloaded_catalogues/{file}'))


def download_vizier_catalog(catalog_keyword: str, columns: list[str] = ['**']) -> pd.DataFrame:
    """Download whole catalog from Vizier with specified columns."""
    data = Vizier(columns=columns,
                  catalog=catalog_keyword,
                  row_limit=-1).get_catalogs(catalog=catalog_keyword)[0]
    data = data.to_pandas()
    return data


from .data_handler import get_pairs_train, get_pairs_predict
from .pipeline import Pipeline
from .connection import Connector


__all__ = ['get_pairs_train', 
           'get_pairs_predict', 
           'Pipeline', 
           'Connector']



import pandas as pd

from typing import List, Tuple, Dict
from cosmatch.utils import add_postfix_to_main_columns, get_attrs_columns
from cosmatch.preprocessing import merge_duplicates_nearest


def _rename_columns(df: pd.DataFrame) -> pd.DataFrame:
    df['id'] = df.attrs['name']+ ': ' +df[df.attrs['id']].astype(str)
    df.rename(columns={df.attrs['ra']: 'ra', df.attrs['dec']: 'dec', df.attrs['poserr']: 'poserr'}, inplace=True)
    return df

def _rename_mags(df: pd.DataFrame, magnitudes_pairs: Dict[str, Tuple[str, str]], num: int) -> pd.DataFrame:
    mag_names = {}
    for key, val in magnitudes_pairs.items():
        if val[num] is not None:
            mag_names[val[num]] = key
    
    df.rename(columns=mag_names, inplace=True)
    df = df[['id', 'ra', 'dec', 'poserr'] +list(mag_names.values())]

    return df

def _concat_plans(df, provided_plan=dict()):
    #provided_plan = {'min': ['u_magErr', 'g_magErr', 'r2_magErr', 'ha_magErr', 'r_magErr', 'i_magErr']}
    plan = {'first': ['ra', 'dec', 'poserr'],
            'sum': ['id']}
    if 'mags' in df.attrs:
        plan['min'] = df.attrs['mags']

    agg = {}
    for key, val in plan.items():
        for i in val:
            agg[i] = key

    for key, val in provided_plan.items():
        for i in val:
            agg[i] = key

    for col in df.columns:
        if col not in agg:
            agg[col] = 'median'

    plan = dict()
    for col, method in agg.items():
        if method not in plan:
            plan[method] = [col]
        else:
            plan[method].append(col)
    return plan


def unite_catalogues(df1: pd.DataFrame, df2:pd.DataFrame, name: str, columns_pairs: Dict[str, Tuple[str, str]] = {},
                     magnitudes: list = [], agg_functions: Dict[str, List[str]] = {},
                     max_distance: float=5, percentile: float=0.99) -> pd.DataFrame:
    df1 = _rename_columns(df1.copy())
    df2 = _rename_columns(df2.copy())

    df1 = _rename_mags(df1, columns_pairs, 0)
    df2 = _rename_mags(df2, columns_pairs, 1)

    df = pd.concat([df1, df2]).reset_index(drop=True)
    df.attrs= {'name': name, 
               'id': 'id', 
               'ra': 'ra', 
               'dec': 'dec', 
               'poserr': 'poserr',
               'mags': list(magnitudes)}

    agg_functions = _concat_plans(df, agg_functions)
    df = merge_duplicates_nearest(df,
                                  max_distance=max_distance, 
                                  percentile=percentile,
                                  verbose=False,
                                  agg_functions=agg_functions)

    add_postfix_to_main_columns(df, postfix=name, add_to_attrs=True)
    df.rename(columns={'poserr': f'poserr_{name}'}, inplace=True)
    df.attrs['poserr'] = f'poserr_{name}'
    return df









from .connection import connection_table_nearest
from ..utils import move_main_columns_ahead

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

# legacy
def _make_union_table(df1: pd.DataFrame, df2: pd.DataFrame, max_distance=1) -> pd.DataFrame:
    """
    Union of two tables with removing same objects.
    """
    id1 = df1.attrs['id']
    id2 = df2.attrs['id']
    cor = connection_table_nearest(df1, df2, max_distance, keep_distance=False)
    df1_out = df1[~df1[id1].isin(cor[id1])][[id1]]
    df1_out[id2] = np.NaN
    df2_out = df2[~df2[id2].isin(cor[id2])][[id2]]
    df2_out[id1] = np.NaN
    cor = pd.concat([cor, df1_out, df2_out])
    return cor

# legacy
def _join_union_table(union_table, df_main, df_support):
    union_table = pd.merge(union_table, df_main, on=df_main.attrs['id'], how='left')
    union_table = pd.merge(union_table, df_support, on=df_support.attrs['id'], how='left')

    union_table['id_union'] = f"{df_main.attrs['name']}: " + union_table['id_ukidss'].astype(str) +\
          f" | {df_support.attrs['name']}: " + union_table['id_mass'].astype(str)
    union_table['ra_union'] = union_table[df_main.attrs['ra']]
    union_table.loc[union_table['ra_union'].isna(), 'ra_union'] = union_table.loc[union_table['ra_union'].isna(), df_support.attrs['ra']]
    union_table['dec_union'] = union_table[df_main.attrs['dec']]
    union_table.loc[union_table['dec_union'].isna(), 'dec_union'] = union_table.loc[union_table['dec_union'].isna(), df_support.attrs['dec']]

    union_table.attrs['id'] = 'id_union'
    union_table.attrs['ra'] = 'ra_union'
    union_table.attrs['dec'] = 'dec_union'
    union_table.attrs['name'] = 'union'

    union_table.drop(columns=[df_main.attrs['id'], df_main.attrs['ra'], df_main.attrs['dec'], 
                              df_support.attrs['id'], df_support.attrs['ra'], df_support.attrs['dec']], inplace=True)
    union_table = move_main_columns_ahead(union_table)
    return union_table

# legacy
def _merge_columns_union_table(union_table: pd.DataFrame, columns_to_merge: Dict[str, Tuple[str, str]],
                              method: str='left', drop: bool=True) -> pd.DataFrame:
    # method = ['left', 'right', 'max', 'min', 'avg']
    for col_res, (col1, col2) in columns_to_merge.items():
        if method=='left':
            union_table[col_res] = union_table[col1].fillna(union_table[col2])
        elif method=='right':
            union_table[col_res] = union_table[col2].fillna(union_table[col1])
        elif method=='max':
            union_table[col_res] = union_table[[col1, col2]].max(axis=1)
        elif method=='min':
            union_table[col_res] = union_table[[col1, col2]].min(axis=1)
        elif method=='avg':
            union_table[col_res] = union_table[[col1, col2]].mean(axis=1)
        if drop: del union_table[col1], union_table[col2]
    return union_table

# legacy
def _plot_union_result(union_table, columns_to_merge): # TODO
    fig, ax = plt.subplots(1, 3,figsize=(10, 3))

    ax[0].scatter(union_table['h_mag_ukidss'], union_table['h_mag_mass'], s=0.01)
    ax[0].plot([10, 18], [10, 18], 'r')
    ax[0].set_xlim(10, 18)
    ax[0].set_ylim(10, 18)
    ax[0].set_title('H-mag')

    ax[1].scatter(union_table['j_mag_ukidss'], union_table['j_mag_mass'], s=0.01)
    ax[1].plot([10, 18], [10, 18], 'r')
    ax[1].set_xlim(10, 18)
    ax[1].set_ylim(10, 18)
    ax[1].set_title('J-mag')

    ax[2].scatter(union_table['k_mag_ukidss'], union_table['k_mag_mass'], s=0.01)
    ax[2].plot([10, 18], [10, 18], 'r')
    ax[2].set_xlim(10, 18)
    ax[2].set_ylim(10, 18)
    ax[2].set_title('K-mag')


"""Future module for classification of sources to different classes."""

import pandas as pd
import numpy as np
from catboost import CatBoostClassifier, Pool
from sklearn.model_selection import train_test_split

from ..utils import correlate
from ..utils import get_attrs_columns
from ..match.connection import Connector

from ..preprocessing import merge_duplicates_same_coords


def _add_classes(data: pd.DataFrame, ignored: list, path_to_class: str = 'DS/LS/LS_class.pickle') -> pd.DataFrame:
    """Initial version of classification."""
    lis = pd.read_pickle(path_to_class)
    classes = correlate(data, lis, 0.3, ('ra_gaia', 'dec_gaia'), ('ra', 'dec'), add_distance=False)
    classes.drop(columns=['id', 'ra', 'dec'], inplace=True)

    X_train, X_test, y_train, y_test = train_test_split(classes.drop(columns=['spectype']), classes['spectype'], stratify=classes['spectype'], test_size=0.2)

    pool_train = Pool(X_train.drop(columns=ignored), y_train)
    pool_test = Pool(X_test.drop(columns=ignored), y_test)

    model = CatBoostClassifier(iterations=1000, depth=6, learning_rate=0.3,
                               loss_function='MultiClass',
                               eval_metric='TotalF1', early_stopping_rounds=50)
    model.fit(pool_train, eval_set=pool_test, verbose=False)

    predict = model.predict_proba(data.drop(columns=ignored))
    data['p_quasar'] = predict[:, 0]
    data['p_galaxy'] = predict[:, 1]
    data['p_star'] = predict[:, 2]
    return data


def add_classes(df_main: pd.DataFrame, class_table: pd.DataFrame, class_column: str, 
                max_distance: float = 5, percentile=0.99, verbose=False, verbose_hist=False) -> pd.DataFrame:
    """
    Add classes from class dataset
    
    Args:
        df_main: main dataframe with sources. Attrs: ['id', 'ra', 'dec', 'poserr']
        class_table: dataframe with classes and its coordinates. Attrs: ['name', 'id', 'ra', 'dec']
        class_column: name of column with classes in class_table.
        max_distance: maximum distance between sources and object in class_table.
    
    Returns:
        df_main with new column ['class_{name}'], where name = class_table.attrs['name'].

    Note:
        Function change input dataframe - add new column: ['class_{name}'].
    """
    df = df_main[get_attrs_columns(df_main, ['id', 'ra', 'dec', 'poserr'])].copy()
    class_table = class_table[get_attrs_columns(class_table, ['id', 'ra', 'dec', 'poserr']) + [class_column]].copy()
    class_table = class_table.drop_duplicates().reset_index(drop=True)
    class_table.attrs['mags'] = []
    class_table = merge_duplicates_same_coords(class_table, agg_functions={'first': [class_column, class_table.attrs['id']]})

    connector = Connector()
    linktable = connector.get_linktable('nearest', df, class_table, max_distance, kwargs={'one_to_one': True, 'percentile': percentile})
    if verbose:
        print('Максимальная дистанция в поиске классов:', linktable[f'distance_to_{class_table.attrs["name"]}'].max())
        print('Медианная дистанция в поиске классов:', linktable[f'distance_to_{class_table.attrs["name"]}'].median())
    if verbose_hist:
        linktable[f'distance_to_{class_table.attrs["name"]}'].hist(bins=100)
    cor = connector.connect_all_linktables([linktable], [class_table, df], df.attrs['id'])
    new_class_column_name = f'{class_column}_{class_table.attrs["name"]}'
    cor = cor[[df.attrs['id'], new_class_column_name]]
    df = pd.merge(df, cor, on=df.attrs['id'], how='left')
    df_main[new_class_column_name] = df[new_class_column_name].values

    return df_main



import os
import requests # type: ignore
import json
import urllib
import pandas as pd
from ..utils import local_package_path

from typing import Union


# Переодически бывают ошибки в скачивании данных. Надо пробовать несколько раз

class DiskLoader:
    """Class for loading catalogs from yandex disk."""

    def __init__(self, file_name: str) -> None:
        """Init."""
        self.folder_link = 'https://disk.yandex.ru/d/dxNUCwXCW5_mHg'
        self.path_to_save = local_package_path(f'downloaded_catalogues/{file_name.replace("/","_")}')
        self.file_name = file_name
        self.columns: Union[dict, None] = None
        self.str_about: Union[str, None] = None

    def _load_file(self) -> pd.DataFrame:
        """Download file."""
        url = 'https://cloud-api.yandex.net/v1/disk/public/resources/download' + '?public_key='\
              + urllib.parse.quote(self.folder_link) + '&path=/' + urllib.parse.quote(self.file_name)

        r = requests.get(url)  # запрос ссылки на скачивание
        h = json.loads(r.text)['href']  # 'парсинг' ссылки на скачивание
        return pd.read_pickle(h)

    def remove(self) -> None:
        """Delete file from local environment."""
        if os.path.exists(self.path_to_save):
            print(f'File {self.path_to_save} already exists. Removing...')
            os.remove(self.path_to_save)

    def load(self, force: bool = False, save: bool = True) -> pd.DataFrame:
        """Download file if in local environment it does not exist."""
        if os.path.exists(self.path_to_save) and not force:
            print(f'File {self.path_to_save} already exists. Loading...')
            return pd.read_pickle(self.path_to_save)

        print(f'File {self.path_to_save} does not exist. Downloading...')
        df = self._load_file()
        if save: df.to_pickle(self.path_to_save)

        return df

    def about(self) -> str:
        """Get description of columns. If you haven't implemented self.about_columns, it will return 'About was not implemented yet."""
        if self.str_about is None:
            return "About was not implemented yet."
        return self.str_about
    
class CSC2(DiskLoader):
    """Class for download CSC2 as support catalogue to teach model of correlation."""

    def __init__(self) -> None:
        """Init."""
        super().__init__('CSC2.pkl')


class DESI(DiskLoader):
    """Class for download DESI LIS, in which DESI classes are stored. Needed for teach model of classification."""

    def __init__(self) -> None:
        """Init."""
        super().__init__('DESI_classes.pkl')



import pickle
import shutil
import os
import requests
import matplotlib.pyplot as plt
import pandas as pd
from .find_hmxb import plot_hmxb

def save_results_to_yandex(tsne_results, data, name):
    

    URL = 'https://cloud-api.yandex.net/v1/disk/resources'
    with open(f'DS/new_hmxb/yandex_token.pickle', 'rb') as f:
        TOKEN = pickle.load(f)
    headers = {'Authorization': f'OAuth {TOKEN}', 'Content-Type': 'application/jpg'}
    folder_path = f'astronomy/blog/result_tsne/{name}'

    def upload_file(loadfile, savefile, replace=False):
        """Загрузка файла.
        savefile: Путь к файлу на Диске
        loadfile: Путь к загружаемому файлу
        replace: true or false Замена файла на Диске"""
        res = requests.get(f'{URL}/upload?path={savefile}&overwrite={replace}', headers=headers).json()
        with open(loadfile, 'rb') as f:
            try:
                requests.put(res['href'], files={'file':f})
            except KeyError:
                print(res)

    def create_folder(path):
        """Создание папки. \n path: Путь к создаваемой папке."""
        requests.put(f'{URL}?path={path}', headers=headers)

    shutil.rmtree('tmp')
    os.mkdir('tmp')

    create_folder(folder_path)
    for key in tsne_results:
        plot_hmxb(tsne_results[key], data, name=f'{name}, perplexity='+str(key))
        path = f"tmp/{name.replace('+', '_')}_{key}.jpg"
        plt.savefig(path)
        upload_file(path, f'{folder_path}/{name}_{key}.jpg', replace=True)

    data.to_csv('tmp/data.csv', index=False)
    upload_file('tmp/data.csv', f'{folder_path}/data.csv', replace=True)

def save_results_to_local(data, results, name):
    os.makedirs(f'DS/new_hmxb/result_tsne/{name}', exist_ok=True)
    data.to_pickle(f'DS/new_hmxb/result_tsne/{name}/data.pickle')
    with open(f'DS/new_hmxb/result_tsne/{name}/results.pickle', 'wb') as f:
        pickle.dump(results, f)

def load_results_from_local(name):
    data = pd.read_pickle(f'DS/new_hmxb/result_tsne/{name}/data.pickle')
    with open(f'DS/new_hmxb/result_tsne/{name}/results.pickle', 'rb') as f:
        results = pickle.load(f)
    return data, results


from .neighbours_loader import NeighboursLoader
from .neighbours_loader import PS1Neighbours
from .neighbours_loader import GaiaDr3Neighbours
from .neighbours_loader import TwoMassNeighbours
from .neighbours_loader import AllWISENeighbours
from .neighbours_loader import CatWISENeighbours
from .neighbours_loader import GLIMPSENeighbours
from .neighbours_loader import UKIDSSNeighbours
from .neighbours_loader import SDSSNeighbours
from .neighbours_loader import SimbadNeighbours
from .neighbours_loader import IPHASNeighbours
from .neighbours_loader import VPHASNeighbours
from .neighbours_loader import NOMADNeighbours


from .disk_loader import DiskLoader
from .disk_loader import CSC2, DESI

from .utils import clear_downloaded_files, download_vizier_catalog

from ..settings import local_package_path
import os

if not os.path.exists(local_package_path('downloaded_catalogues')):
    os.makedirs(local_package_path('downloaded_catalogues'))

__all__ = [
    'NeighboursLoader',
    'PS1Neighbours',
    'GaiaDr3Neighbours',
    'TwoMassNeighbours',
    'AllWISENeighbours',
    'CatWISENeighbours',
    'GLIMPSENeighbours',
    'UKIDSSNeighbours',
    'SDSSNeighbours',
    'SimbadNeighbours',
    'IPHASNeighbours',
    'VPHASNeighbours',
    'NOMADNeighbours',

    'DiskLoader',
    'CSC2',
    'DESI',

    'clear_downloaded_files',
    'download_vizier_catalog',
]



"""Modul for loading data from Vizier."""

from astroquery.vizier import Vizier
import astropy.coordinates as coord
import astropy.units as u

import pandas as pd
import numpy as np
import requests
import warnings
import re
import shutil
import os

from abc import ABC, abstractmethod

from ..utils import local_package_path, add_postfix_to_main_columns, add_magnitudes_to_attrs

from typing import Union


class NeighboursLoader(ABC):
    """Abstract class for finding near pairs of object from CDS X-Match http://cdsxmatch.u-strasbg.fr.

    To implement this class, you need to define the following methods: __init__

    In this method you should:
        path - set the path to the catalog from vizier like 'vizier:{catalog_name}', where catalog_name from vizier (e.g. 'II/349/ps1')
        name - set the name of the catalog, it will be used in name of columns ['id', 'ra', 'dec'] and in attrs
    Optional:
        rename_columns: dict[str, str] - rename columns in data from VizieR. It is strongly recommended to make ['id', 'ra', 'dec'] in columns,\
            since it will be renamed to ['id_{name}', 'ra_{name}', 'dec_{name}'] and this columns will be added in attrs.
        keep_columns: list[str] - what columns to keep in dataset. If it is None, all columns will be kept.
        magnitudes: list[str] - list of all names of magnitudes after renaming. This columns will be added in attrs['magnitudes'].
        filter_query: str - filter query for VizieR. It will be used to delete some object, with which statement is not true.
    """

    def __init__(self) -> None:
        """Initialize the class. You should implement this method in your class."""
        self.path: str  # name of catalog in Vizier (like )
        self.name: str
        self.filter_query: Union[str, None]= None
        self.rename_columns: Union[dict[str, str], None] = None
        self.keep_columns: Union[list[str], None] = None
        self.magnitudes: Union[list[str], None] = None
        self.dtypes: Union[dict[str, str], None] = None

    def _save_part(self, sources: pd.DataFrame, num: int, window: int, folder_path: str) -> None:
        """Save part of sources."""
        part = sources[num * window:(num + 1) * window]
        part = part[[sources.attrs['id'], sources.attrs['ra'], sources.attrs['dec']]]
        part.to_csv(f'{folder_path}/part.csv', index=False)

    def _request_part(self, num: int, folder_path: str, attrs: dict, max_distance: float) -> None:
        """Request data from CDS X-Match."""
        path_to_part = f'{folder_path}/part.csv'
        path_to_save_csv = f'{folder_path}/{num}.csv'
        path_to_save_pkl = f'{folder_path}/{num}.pkl'

        r = requests.post(
            'http://cdsxmatch.u-strasbg.fr/xmatch/api/v1/sync',
            data={'request': 'xmatch', 'distMaxArcsec': max_distance, 'RESPONSEFORMAT': 'csv',
            'cat2': self.path, 'colRA1': attrs['ra'], 'colDec1': attrs['dec']},
            files={'cat1': open(path_to_part, 'r')})
        if r.status_code != 200:
            # Extract the value of the QUERY_STATUS tag
            query_status_match = re.search(r'<INFO name="QUERY_STATUS" value="([^"]+)">', r.text)
            if query_status_match:
                query_status_value = query_status_match.group(1)
                print(f"Query status code: {query_status_value}")
            else:
                print("Query status code not found in the response.")
        h = open(path_to_save_csv, 'w')
        h.write(r.text)
        h.close()
        try:
            output = pd.read_csv(path_to_save_csv)
        except BaseException:
            with open(path_to_save_csv, 'r') as f:
                response = f.read()
                error_message_pattern = r'<INFO name="QUERY_STATUS" value="ERROR">\s*Message:\s*(.+?)\s*</INFO>'
                error_message = re.search(error_message_pattern, response, re.DOTALL)

                if error_message:
                    error_message = error_message.group(1).strip()
                    raise Exception(f"Error message: {error_message}")
                else:
                    raise Exception("Error message not found in the response.")
        return output
    
    def _handle_part(self, df: pd.DataFrame, attrs: dict) -> pd.DataFrame:
        df.drop(columns=[attrs['id'], attrs['ra'], attrs['dec'], 'angDist'], inplace=True)
        if self.rename_columns: df.rename(columns=self.rename_columns, inplace=True)
        if self.filter_query: df.query(self.filter_query, inplace=True)
        if self.keep_columns: df = df[self.keep_columns].copy()
        df.drop_duplicates(inplace=True)
        df.reset_index(drop=True, inplace=True)
        return df
    
    def _merge_parts(self, folder_path: str, size: int) -> pd.DataFrame:
        ans = pd.DataFrame()
        for i in range(size):
            part = pd.read_pickle(os.path.join(folder_path, f'{i}.pkl'))
            if len(part) == 0: continue
            ans = pd.concat([ans, part])
        return ans
    
    def _change_dtypes(self, df: pd.DataFrame) -> pd.DataFrame:
        if not self.dtypes:
            return df
        for i in self.dtypes:
            df[i] = df[i].astype(self.dtypes[i])
        return df

    def _handle_full(self, df: pd.DataFrame) -> pd.DataFrame:
        df.drop_duplicates(inplace=True)
        if 'id' in df.columns and 'ra' in df.columns and 'dec' in df.columns:
            add_postfix_to_main_columns(df, self.name)
        else:
            warnings.warn(f'Columns id, ra, dec not found in columns after rename. Not adding postfix to columns and to attrs.')
        if 'poserr' in df.columns: 
            df.rename(columns={'poserr': 'poserr_' + self.name}, inplace=True)
            df.attrs['poserr'] = 'poserr_' + self.name
        if self.magnitudes: add_magnitudes_to_attrs(df, self.magnitudes, self.name)
        df = self._change_dtypes(df)
        return df

    def load_data(self, sources: pd.DataFrame, max_distance: float = 10,
                  verbose: bool = False, window: int = 1000, start_at: int = 0) -> pd.DataFrame:
        """
        Load data from CDS X-Match.

        Args:
            sourses: catalog of objects with coordinates [ra, dec]
            radius_arcsec: radius of search in arcsec.
            verbose: if True, print progress.
            window: number of sources to load at once.
        Returns:
            Dataframe with loaded data.
        Note:
            This method generate temporary files for loading data. It also save data.
        """
        path_folder = local_package_path(os.path.join('downloaded_catalogues', 'tmp_neighbours_loader'))

        if not os.path.exists(path_folder):
            os.mkdir(path_folder)

        path_folder = os.path.join(path_folder, self.name)

        if not os.path.exists(path_folder):
            os.mkdir(path_folder)

        for num in range((len(sources)-1) // window + 1):
            if num*window < start_at: continue
            self._save_part(sources, num, window, path_folder)

            if verbose: print(f"{num*window} start loading...", end=' ')
            part = self._request_part(num, path_folder, sources.attrs, max_distance=max_distance)
            if verbose: print(f"Start saving...", end=' ')
            part = self._handle_part(part, sources.attrs)
            part.to_pickle(os.path.join(path_folder, f'{num}.pkl'))
            if verbose: print(f"Finished!", end='\n')

        if verbose: print('Data collected successfully. Start to merge.')
        df = self._merge_parts(path_folder, size=(len(sources)-1) // window + 1)
        df = self._handle_full(df)
        if verbose: print(f'Data merged successfully.')

        for file in os.listdir(path_folder):
            if file !='part.csv':
                os.remove(os.path.join(path_folder, file))
        #shutil.rmtree(path_folder)

        return df


class PS1Neighbours(NeighboursLoader):
    """Class for loading PanSTARRS DR1."""

    def __init__(self) -> None:
        """Init."""
        super().__init__()
        self.path = 'vizier:II/349/ps1'
        self.name = 'ps'

        #self.filter_query = 'Ns > 5 & Nd > 5'
        self.rename_columns = {'objID': 'id', 'RAJ2000': 'ra', 'DEJ2000': 'dec', 'errHalfMaj': 'poserr',
                               'gmag': 'g_mag', 'rmag': 'r_mag', 'imag': 'i_mag', 'zmag': 'z_mag', 'ymag': 'y_mag',
                               'gKmag': 'g_K_mag', 'rKmag': 'r_K_mag', 'iKmag': 'i_K_mag', 'zKmag': 'z_K_mag', 'yKmag': 'y_K_mag',
                               'e_gmag': 'g_magErr', 'e_rmag': 'r_magErr', 'e_imag': 'i_magErr', 'e_zmag': 'z_magErr', 'e_ymag': 'y_magErr',
                               'e_gKmag': 'g_K_magErr', 'e_rKmag': 'r_K_magErr', 'e_iKmag': 'i_K_magErr', 'e_zKmag': 'z_K_magErr', 'e_yKmag': 'y_K_magErr',}
        self.magnitudes = ['g_mag', 'r_mag', 'i_mag', 'z_mag', 'y_mag',
                           'g_K_mag', 'r_K_mag', 'i_K_mag', 'z_K_mag', 'y_K_mag']
        self.keep_columns = ['id', 'ra', 'dec', 'poserr', 
                             'Ns', 'Nd',
                             'g_mag', 'r_mag', 'i_mag', 'z_mag', 'y_mag',
                             'g_magErr', 'r_magErr', 'i_magErr', 'z_magErr', 'y_magErr',
                             'g_K_mag', 'r_K_mag', 'i_K_mag', 'z_K_mag', 'y_K_mag',
                             'g_K_magErr', 'r_K_magErr', 'i_K_magErr', 'z_K_magErr', 'y_K_magErr']
        self.dtypes = {}
        for col in self.keep_columns:
            if col not in ['id', 'ra', 'dec', 'poserr']:
                self.dtypes[col] = 'float32'
        del self.dtypes['Ns'], self.dtypes['Nd']

class GaiaDr3Neighbours(NeighboursLoader):
    """Class for loading GAIA DR3."""

    def __init__(self) -> None:
        """Init."""
        super().__init__()
        self.path = 'vizier:I/355/gaiadr3'
        self.name = 'gaia'

        self.rename_columns = {'Source': 'id', 'RAdeg': 'ra', 'DEdeg': 'dec', 'errHalfMaj': 'poserr',
                               'Plx': 'paralax', 'e_Plx': 'paralaxErr', 
                               'PM': 'pm', 'pmRA': 'pm_ra', 'pmDE': 'pm_dec', 'e_pmRA': 'pm_raErr', 'e_pmDE': 'pm_decErr',
                               'Gmag': 'g_mag', 'BPmag': 'bp_mag', 'RPmag': 'rp_mag',
                               'e_Gmag': 'g_magErr', 'e_BPmag': 'bp_magErr', 'e_RPmag': 'rp_magErr'}
        self.magnitudes = ['g_mag', 'bp_mag', 'rp_mag']
        self.keep_columns = ['id', 'ra', 'dec', 'poserr', 'paralax', 'paralaxErr', 
                             'pm', 'pm_ra', 'pm_dec', 'pm_raErr', 'pm_decErr',
                             'g_mag', 'g_magErr', 'bp_mag', 'bp_magErr', 'rp_mag', 'rp_magErr']
        
        self.dtypes = {}
        for col in self.keep_columns:
            if col not in ['id', 'ra', 'dec', 'poserr']:
                self.dtypes[col] = 'float32'
        
class TwoMassNeighbours(NeighboursLoader):
    """Class for loading 2MASS."""
    
    def __init__(self) -> None:
        """Init."""
        super().__init__()
        self.path = 'vizier:II/246/out'
        self.name = '2mass'

        self.rename_columns = {'2MASS': 'id', 'RAJ2000': 'ra', 'DEJ2000': 'dec', 'errHalfMin': 'poserr',
                               'Jmag': 'j_mag', 'e_Jmag': 'j_magErr', 
                               'Hmag': 'h_mag', 'e_Hmag': 'h_magErr',
                               'Kmag': 'k_mag', 'e_Kmag': 'k_magErr'}
        self.magnitudes = ['j_mag', 'h_mag', 'k_mag']
        self.keep_columns = ['id', 'ra', 'dec', 'poserr', 
                             'j_mag', 'j_magErr', 'h_mag', 'h_magErr', 'k_mag', 'k_magErr']

        self.all_columns = ['2MASS', 'RAJ2000', 'DEJ2000', 'errHalfMaj', 'errHalfMin', 'errPosAng',
                            'Jmag', 'Hmag', 'Kmag', 'e_Jmag', 'e_Hmag', 'e_Kmag', 
                            'Qfl', 'Rfl', 'X', 'MeasureJD']
        
        self.dtypes = {}
        for col in self.keep_columns:
            if col not in ['id', 'ra', 'dec', 'poserr']:
                self.dtypes[col] = 'float32'

    
class AllWISENeighbours(NeighboursLoader):

    def __init__(self) -> None:
        """Init."""
        super().__init__()
        self.path = 'vizier:II/328/allwise'
        self.name = 'allwise'

        self.rename_columns = {'ID': 'id', 'RAJ2000': 'ra', 'DEJ2000': 'dec', 'eeMaj': 'poserr',
                               'W1mag': 'w1_mag', 'W2mag': 'w2_mag', 'W3mag': 'w3_mag', 'W4mag': 'w4_mag',
                               'e_W1mag': 'w1_magErr', 'e_W2mag': 'w2_magErr', 'e_W3mag': 'w3_magErr', 'e_W4mag': 'w4_magErr',
                               'Jmag': 'j_mag', 'Hmag': 'h_mag', 'Kmag': 'k_mag',
                               'e_Jmag': 'j_magErr', 'e_Hmag': 'h_magErr', 'e_Kmag': 'k_magErr',
                               'pmRA': 'pm_ra', 'pmDE': 'pm_dec', 'e_pmRA': 'pm_raErr', 'e_pmDE': 'pm_decErr'}
        self.magnitudes = ['w1_mag', 'w2_mag', 'w3_mag', 'w4_mag', 'j_mag', 'h_mag', 'k_mag']
        self.keep_columns = ['id', 'ra', 'dec', 'poserr', 
                             'w1_mag', 'w2_mag', 'w3_mag', 'w4_mag',
                             'w1_magErr', 'w2_magErr', 'w3_magErr', 'w4_magErr',
                             'j_mag', 'h_mag', 'k_mag',
                             'j_magErr', 'h_magErr', 'k_magErr',
                             'pm_ra', 'pm_raErr', 'pm_dec', 'pm_decErr']
        
        self.all_columns = ['AllWISE', 'RAJ2000', 'DEJ2000', 'eeMaj', 'eeMin', 'eePA', 'W1mag',
                            'W2mag', 'W3mag', 'W4mag', 'Jmag', 'Hmag', 'Kmag', 'e_W1mag', 'e_W2mag',
                            'e_W3mag', 'e_W4mag', 'e_Jmag', 'e_Hmag', 'e_Kmag', 'ID', 'ccf', 'ex',
                            'var', 'qph', 'pmRA', 'e_pmRA', 'pmDE', 'e_pmDE', 'd2M']
        
        self.dtypes = {}
        for col in self.keep_columns:
            if col not in ['id', 'ra', 'dec', 'poserr']:
                self.dtypes[col] = 'float32'
        

class CatWISENeighbours(NeighboursLoader):

    def __init__(self) -> None:
        """Init."""
        super().__init__()
        self.path = 'vizier:II/365/catwise'
        self.name = 'catwise'

        self.rename_columns = {'objID': 'id', 'RAdeg': 'ra', 'DEdeg': 'dec', 'errHalfMaj': 'poserr',
                               'plx1': 'paralax1', 'e_plx1': 'paralax1Err', 'plx2': 'paralax2', 'e_plx2': 'paralax2Err',
                               'W1mproPM': 'w1_mag', 'W2mproPM': 'w2_mag', 'e_W1mproPM': 'w1_magErr', 'e_W2mproPM': 'w2_magErr',
                               'snrW1pm': 'w1_snr', 'snrW2pm': 'w2_snr', 
                               'pmRA': 'pm_ra', 'pmDE': 'pm_dec', 'e_pmRA': 'pm_raErr', 'e_pmDE': 'pm_decErr'}
        self.keep_columns = ['id', 'ra', 'dec', 'poserr', 'paralax1', 'paralax1Err', 'paralax2', 'paralax2Err',
                             'w1_mag', 'w2_mag', 'w1_magErr', 'w2_magErr', 'w1_snr', 'w2_snr',
                             'pm_ra', 'pm_raErr', 'pm_dec', 'pm_decErr']
        self.magnitudes = ['w1_mag', 'w2_mag']

        self.all_columns = ['objID', 'RAdeg', 'DEdeg', 'errHalfMaj', 'errHalfMin', 'errPosAng',
                            'Name', 'e_RAdeg', 'e_DEdeg', 'ePos', 'nW1', 'mW1', 'nW2', 'mW2', 'MJD',
                            'RAPMdeg', 'DEPMdeg', 'e_RAPMdeg', 'e_DEPMdeg', 'ePosPM', 'pmRA',
                            'pmDE', 'e_pmRA', 'e_pmDE', 'snrW1pm', 'snrW2pm', 'FW1pm', 'e_FW1pm',
                            'FW2pm', 'e_FW2pm', 'W1mproPM', 'e_W1mproPM', 'W2mproPM', 'e_W2mproPM',
                            'pmQual', 'Dist', 'chi2pmRA', 'chi2pmDE', 'ka', 'k1', 'k2', 'km',
                            'plx1', 'e_plx1', 'plx2', 'e_plx2', 'Sep', 'ccf', 'abfl']
        
        self.dtypes = {}
        for col in self.keep_columns:
            if col not in ['id', 'ra', 'dec', 'poserr']:
                self.dtypes[col] = 'float32'
        
    
class GLIMPSENeighbours(NeighboursLoader):
    """Class for loading GLIMPSE."""
    
    def __init__(self) -> None:
        """Init."""
        super().__init__()
        self.path = 'vizier:II/293/glimpse'
        self.name = 'glimpse'
        
        self.rename_columns = {'GLIMPSE': 'id', 'RAdeg': 'ra', 'DEdeg': 'dec', 'errHalfMaj': 'poserr',
                               '3.6mag': '3.6_mag', '4.5mag': '4.5_mag', '5.8mag': '5.8_mag', '8.0mag': '8.0_mag',
                               'e_3.6mag': '3.6_magErr', 'e_4.5mag': '4.5_magErr', 'e_5.8mag': '5.8_magErr', 'e_8.0mag': '8.0_magErr',}
        self.magnitudes = ['3.6_mag', '4.5_mag', '5.8_mag', '8.0_mag']
        self.keep_columns = ['id', 'ra', 'dec', 'poserr', 
                             '3.6_mag', '4.5_mag', '5.8_mag', '8.0_mag',
                             '3.6_magErr', '4.5_magErr', '5.8_magErr', '8.0_magErr']

        self.all_columns = ['GLIMPSE', 'RAdeg', 'DEdeg', 'errHalfMaj', 'errHalfMin', 'errPosAng',
                            '3.6mag', '4.5mag', '5.8mag', '8.0mag', 'e_3.6mag', 'e_4.5mag',
                            'e_5.8mag', 'e_8.0mag', 'C', 'F(3.6)', 'e_F(3.6)', 'F(4.5)', 'e_F(4.5)',
                            'F(5.8)', 'e_F(5.8)', 'F(8.0)', 'e_F(8.0)', 'q_3.6mag', 'q_4.5mag',
                            'q_5.8mag', 'q_8.0mag']
        
        self.dtypes = {}
        for col in self.keep_columns:
            if col not in ['id', 'ra', 'dec', 'poserr']:
                self.dtypes[col] = 'float32'

                    
class UKIDSSNeighbours(NeighboursLoader):

    def __init__(self) -> None:
        super().__init__()
        self.path = 'vizier:II/316/gps6' # gcs9 , las9
        self.name = 'ukidss'

        self.rename_columns = {'sourceID': 'id', 'ra': 'ra', 'dec': 'dec',
                               'jAperMag3': 'j_mag', 'hAperMag3': 'h_mag', 'k_1AperMag3': 'k_mag',
                               'jAperMag3Err': 'j_magErr', 'hAperMag3Err': 'h_magErr', 'k_1AperMag3Err': 'k_magErr'}
        self.filter_query = 'mode == 1'
        self.magnitudes = ['j_mag', 'h_mag', 'k_mag']
        self.keep_columns = ['id', 'ra', 'dec', 
                             'j_mag', 'h_mag', 'k_mag',
                             'j_magErr', 'h_magErr', 'k_magErr']                      

        self.all_columns = ['JName', 'ra', 'dec', 'jAperMag3', 'hAperMag3', 'k_1AperMag3',
                            'k_2AperMag3', 'h2AperMag3', 'jAperMag3Err', 'hAperMag3Err',
                            'k_1AperMag3Err', 'k_2AperMag3Err', 'h2AperMag3.1', 'sourceID', 'mode',
                            'epoch', 'mergedClass']
        
        self.dtypes = {}
        for col in self.keep_columns:
            if col not in ['id', 'ra', 'dec', 'poserr']:
                self.dtypes[col] = 'float32'
        

class SDSSNeighbours(NeighboursLoader):
    """Class for loading SDSS."""

    def __init__(self) -> None:
        """Init."""
        super().__init__()
        self.path = 'vizier:V/154/sdss16'
        self.name = 'sdss'

        self.rename_columns = {'objID': 'id', 'RA_ICRS': 'ra', 'DE_ICRS': 'dec', 'errHalfMaj': 'poserr',
                               'umag': 'u_mag', 'gmag': 'g_mag', 'rmag': 'r_mag', 'imag': 'i_mag', 'zmag': 'z_mag',
                               'e_umag': 'u_magErr', 'e_gmag': 'g_magErr', 'e_rmag': 'r_magErr', 'e_imag': 'i_magErr',
                               'e_zmag': 'z_magErr', 'pmRA': 'pm_ra', 'pmDE': 'pm_dec',
                               'e_pmRA': 'pm_raErr', 'e_pmDE': 'pm_decErr'}
        self.magnitudes = ['u_mag', 'g_mag', 'r_mag', 'i_mag', 'z_mag']
        self.keep_columns = ['id', 'ra', 'dec', 'poserr', 
                             'u_mag', 'g_mag', 'r_mag', 'i_mag', 'z_mag',
                             'u_magErr', 'g_magErr', 'r_magErr', 'i_magErr', 'z_magErr',
                             'pm_ra', 'pm_dec', 'pm_raErr', 'pm_decErr']
        self.filter_query = 'mode==1'

        self.all_columns = ['objID', 'RA_ICRS', 'DE_ICRS', 'errHalfMaj', 'errHalfMin', 'errPosAng',
                            'mode', 'class', 'clean', 'e_RA_ICRS', 'e_DE_ICRS', 'umag', 'gmag',
                            'rmag', 'imag', 'zmag', 'e_umag', 'e_gmag', 'e_rmag', 'e_imag',
                            'e_zmag', 'zsp', 'e_zsp', 'f_zsp', 'spCl', 'subCl', 'pmRA', 'pmDE',
                            'e_pmRA', 'e_pmDE', 'sigRA', 'sigDE', 'zph', 'e_zph', '<zph>', 'Q',
                            'SDSS16', 'Sp-ID', 'MJD']
        
        self.dtypes = {}
        for col in self.keep_columns:
            if col not in ['id', 'ra', 'dec', 'poserr']:
                self.dtypes[col] = 'float32'
        

class SimbadNeighbours(NeighboursLoader):
    """Class for loading Simbad."""

    def __init__(self) -> None:
        """Init."""
        super().__init__()
        self.path = 'simbad'
        self.name = 'simbad'

        self.rename_columns = {'main_id': 'id', 'ra': 'ra', 'DE_ICRS': 'dec', 'coo_err_maj': 'poserr',
                               'main_type': 'object_type', 'other_types': 'object_type_other',
                               'radvel': 'velocity', 'radvel_err': 'velocity_err', 
                               'redshift': 'redshift', 'redshift_err': 'redshift_err',
                               'plx': 'paralax', 'plx_err': 'paralax_err',
                               'pmra': 'pm_ra', 'pmdec': 'pm_dec',
                               'B': 'b_mag', 'V': 'v_mag', 'R': 'r_mag', 'J': 'j_mag', 'H': 'h_mag', 'K': 'k_mag',
                               'u': 'u_mag', 'g': 'g_mag', 'r': 'r_mag', 'i': 'i_mag', 'z': 'z_mag'}
        self.magnitudes = ['b_mag', 'v_mag', 'r_mag', 'j_mag', 'h_mag', 'k_mag', 'u_mag', 'g_mag', 'r_mag', 'i_mag', 'z_mag']
        self.keep_columns = ['id', 'ra', 'dec', 'poserr',
                             'object_type', 'object_type_other',
                             'velocity', 'velocity_err',
                             'redshift', 'redshift_err',
                             'paralax', 'paralax_err',
                             'pm_ra', 'pm_dec',
                             'b_mag', 'v_mag', 'r_mag', 'j_mag', 'h_mag', 'k_mag',
                             'u_mag', 'g_mag', 'r_mag', 'i_mag', 'z_mag']
        
        self.all_columns = ['main_id', 'ra', 'dec', 'coo_err_maj', 'coo_err_min', 'coo_err_angle',
                            'nbref', 'ra_sexa', 'dec_sexa', 'coo_qual', 'coo_bibcode', 'main_type',
                            'other_types', 'radvel', 'radvel_err', 'redshift', 'redshift_err',
                            'sp_type', 'morph_type', 'plx', 'plx_err', 'pmra', 'pmdec',
                            'pm_err_maj', 'pm_err_min', 'pm_err_pa', 'size_maj', 'size_min',
                            'size_angle', 'B', 'V', 'R', 'J', 'H', 'K', 'u', 'g', 'r', 'i', 'z']
        
        self.dtypes = {}
        for col in self.keep_columns:
            if col not in ['id', 'ra', 'dec', 'poserr']:
                self.dtypes[col] = 'float32'
        del self.dtypes['object_type'], self.dtypes['object_type_other']


class IPHASNeighbours(NeighboursLoader):
    """Class for loading IPHAS."""

    def __init__(self) -> None:
        """Init."""
        super().__init__()
        self.path = 'vizier:II/321/iphas2'
        self.name = 'iphas'

        self.rename_columns = {'name': 'id', 'ra': 'ra', 'dec': 'dec', 'radec_err': 'poserr',
                               'r': 'r_mag', 'i': 'i_mag', 'ha': 'ha_mag',
                               'rErr': 'r_magErr', 'iErr': 'i_magErr', 'haErr': 'ha_magErr'}
        self.magnitudes = ['r_mag', 'i_mag', 'ha_mag']
        self.filter_query = 'a10point==1'
        self.keep_columns = ['id', 'ra', 'dec', 'poserr',
                             'r_mag', 'i_mag', 'ha_mag',
                             'r_magErr', 'i_magErr', 'ha_magErr']
        
        self.all_columns = ['name', 'ra', 'dec', 'radec_err', 'r', 'rErr', 'i', 'iErr', 'ha',
                            'haErr', 'a10', 'a10point']
        
        self.dtypes = {}
        for col in self.keep_columns:
            if col not in ['id', 'ra', 'dec', 'poserr']:
                self.dtypes[col] = 'float32'
        
    
class VPHASNeighbours(NeighboursLoader):
    """
    Class for loading VPHAS from VizieR
    """

    def __init__(self) -> None:
        """Init."""
        super().__init__()
        self.path = 'vizier:II/341/vphasp'
        self.name = 'vphas'

        self.rename_columns = {'sourceID': 'id', 'RAJ2000': 'ra', 'DEJ2000': 'dec',
                               'umag': 'u_mag', 'gmag': 'g_mag', 'r2mag': 'r2_mag', 'Hamag': 'ha_mag', 
                               'rmag': 'r_mag', 'imag': 'i_mag',
                               'e_umag': 'u_magErr', 'e_gmag': 'g_magErr', 'e_r2mag': 'r2_magErr', 'e_Hamag': 'ha_magErr', 
                               'e_rmag': 'r_magErr', 'e_imag': 'i_magErr',
                               }
        self.keep_columns = ['id', 'ra', 'dec', 'u_mag', 'g_mag', 'r2_mag', 'ha_mag', 'r_mag', 'i_mag',
                             'u_magErr', 'g_magErr', 'r2_magErr', 'ha_magErr', 'r_magErr', 'i_magErr']
        self.magnitudes = ['u_mag', 'g_mag', 'r2_mag', 'ha_mag', 'r_mag', 'i_mag']

        self.all_columns = ['sourceID', 'RAJ2000', 'DEJ2000', 'VPHASDR2', 'clean', 'cleanu', 'umag',
                            'e_umag', 'MJDu', 'cleang', 'gmag', 'e_gmag', 'MJDg', 'cleanr2',
                            'r2mag', 'e_r2mag', 'MJDr2', 'cleanha', 'Hamag', 'e_Hamag', 'MJDha',
                            'cleanr', 'rmag', 'e_rmag', 'MJDr', 'cleani', 'imag', 'e_imag', 'MJDi']
        
        self.dtypes = {}
        for col in self.keep_columns:
            if col not in ['id', 'ra', 'dec', 'poserr']:
                self.dtypes[col] = 'float32'
        

class NOMADNeighbours(NeighboursLoader):
    """Class for loading NOMAD."""

    def __init__(self) -> None:
        """Init."""
        super().__init__()
        self.path = 'vizier:I/297/out'
        self.name = 'nomad'
        
        self.rename_columns = {'NOMAD1.0': 'id', 'RAJ2000': 'ra', 'DEJ2000': 'dec', 'errHalfMaj': 'poserr',
                               'Bmag': 'b_mag', 'Vmag': 'v_mag', 'Rmag': 'r_mag', 
                               'Jmag': 'j_mag', 'Hmag': 'h_mag', 'Kmag': 'k_mag',
                               'pmRA': 'pm_ra', 'pmDE': 'pm_dec'}
        self.magnitudes = ['b_mag', 'v_mag', 'r_mag', 'j_mag', 'h_mag', 'k_mag']
        self.keep_columns = ['id', 'ra', 'dec', 'poserr', 
                             'b_mag', 'v_mag', 'r_mag', 'j_mag', 'h_mag', 'k_mag', 
                             'pm_ra', 'pm_dec']

        self.all_columns = ['NOMAD1.0', 'RAJ2000', 'DEJ2000', 'errHalfMaj', 'errHalfMin',
                            'errPosAng', 'Bmag', 'Vmag', 'Rmag', 'Jmag', 'Hmag', 'Kmag', 'YM', 'r',
                            'pmRA', 'pmDE', 'spRA', 'spDE']
        
        self.dtypes = {}
        for col in self.keep_columns:
            if col not in ['id', 'ra', 'dec', 'poserr']:
                self.dtypes[col] = 'float32'
        

class XMMNeighbours(NeighboursLoader):

    def __init__(self) -> None:
        super().__init__()
        self.path = 'vizier:IX/69/xmm4d13s'
        self.name = 'xmm'


"""Module for correlation several dataframes to one frame."""

import pandas as pd
import numpy as np

from ..utils import correlate, keep_nearest_pairs, merge_save_attrs, move_main_columns_ahead, get_attrs_columns
from ..utils import check_attrs_structure, move_columns_ahead, add_postfix_to_all_columns_except

from ..match import get_pairs_train, get_pairs_predict, Pipeline
from ..transforms import AttrsAutoTransform
from ..match.models import Model

from typing import List

# Lagacy
def connection_table_nearest(df_main: pd.DataFrame, df_sub: pd.DataFrame, max_dist: float = 0.5,
                             keep_distance: bool = True, percentile: float = 0.98) -> pd.DataFrame:
    """
    Correlate two dataframes and keep nearest pairs to all sources in df_main.

    Args:
        df_main: Main dataframe with sources.
        df_sub: Sub dataframe with sources.
        max_dist: Maximum distance between sources in df_main and df_sub.
        keep_distance: Keep distance between sources in df_main and df_sub in the output.

    Returns:
        Correlated dataframe with columns id_main, id_sub and (optional) distance. id_main is unique.
    """
    id_main, ra_main, dec_main = df_main.attrs['id'], df_main.attrs['ra'], df_main.attrs['dec']
    id_sub, ra_sub, dec_sub = df_sub.attrs['id'], df_sub.attrs['ra'], df_sub.attrs['dec']
    df_cor = correlate(df_main, df_sub, max_dist, (ra_main, dec_main), (ra_sub, dec_sub), add_distance=True, add_attrs=False)
    df_cor = keep_nearest_pairs(df_cor, id_main)
    # df_cor = keep_nearest_pairs(df_cor, id_main)
    col = [id_main, id_sub]
    if keep_distance:
        col.append('distance')
    df_cor.attrs['ids'] = [id_main, id_sub]
    return df_cor[col]

class Connector:

    def __init__(self):
        pass

    def _check_attrs(self, df_main: pd.DataFrame, df_sub: pd.DataFrame) -> None:
        check_attrs_structure(df_main, ['name', 'id', 'ra', 'dec', 'poserr'])
        check_attrs_structure(df_sub, ['name', 'id', 'ra', 'dec', 'poserr'])

    def _check_kwargs(self, kwargs: dict, keys: dict) -> None:
        for key in keys:
            if key not in kwargs:
                raise ValueError(f'{key} not in kwargs to connect method.')
            if not isinstance(kwargs[key], keys[key]):
                raise ValueError(f'{key} must be {keys[key]}.')
            
    def _check_duplicates(self, df: pd.DataFrame) -> None:
        if len(df) != len(df.drop_duplicates()):
            raise ValueError('Duplicated rows are in dataset. Use df.drop_duplicates() to remove them.')
        if not df[df.attrs['id']].is_unique:
            raise ValueError('Id is not unique. Use merge_duplicates_same_id from cosmatch.preprocessing to remove them.')
        if df[[df.attrs['ra'], df.attrs['dec']]].duplicated().any():
            raise ValueError('Coordinates are duplicated. Use merge_duplicates_same_coords from cosmatch.preprocessing to remove them.')

    def _connect_nearest(self, df_main: pd.DataFrame, df_sub: pd.DataFrame, max_distance: float,
                         kwargs: dict = {}) -> pd.DataFrame:
        self._check_kwargs(kwargs, {'percentile': float, 'one_to_one': bool})

        percentile = kwargs['percentile']
        one_to_one = kwargs['one_to_one']

        id_main, poserr_main = get_attrs_columns(df_main, ['id', 'poserr'])
        id_sub, poserr_sub = get_attrs_columns(df_sub, ['id', 'poserr'])

        data = correlate(df_main, df_sub, max_distance, add_distance=True, add_attrs=False)
        data = keep_nearest_pairs(data, id_main)
        
        if one_to_one: data = keep_nearest_pairs(data, id_sub)

        if percentile<1 and percentile>0:
            data['percentile_main'] = (-2 * (data[poserr_main]**2) * np.log(1 - percentile))**0.5
            data['percentile_sub'] = (-2 * (data[poserr_sub]**2) * np.log(1 - percentile))**0.5
            data.query(f'distance < percentile_main + percentile_sub', inplace=True)

        col = [id_main, id_sub, f'distance_to_{df_sub.attrs["name"]}']
        data.attrs['ids'] = [id_main, id_sub]
        data.rename(columns={'distance': f'distance_to_{df_sub.attrs["name"]}'}, inplace=True)
        return data[col]
    
    def _connect_trainable_model(self, df_main: pd.DataFrame, df_sub: pd.DataFrame, max_distance: float,
                        kwargs: dict = {}) -> pd.DataFrame:
        self._check_kwargs(kwargs, {'model': Model, 'one_to_one': bool, 'catalog_support': pd.DataFrame, 'P_0_lim': float, 'P_i_lim': float})
        if not kwargs['model'].fit_required:
            raise ValueError('Model does not require training. Use `non-trainable model` method instead.')
        model = kwargs['model']
        one_to_one = kwargs['one_to_one']
        catalog_support = kwargs['catalog_support']
        P_0_lim = kwargs['P_0_lim']
        P_i_lim = kwargs['P_i_lim']

        train_pairs = get_pairs_train(df_main, df_sub, catalog_support, max_distance=max_distance)
        predict_pairs = get_pairs_predict(df_main, df_sub, max_distance=max_distance)
        print(train_pairs['mark'].sum(), len(train_pairs), len(predict_pairs))
        #return train_pairs, predict_pairs

        transform = AttrsAutoTransform(max_distance=max_distance).get_transforms(train_pairs)
        transform.transform(train_pairs, {'poserr': train_pairs.attrs['poserr'], 
                                          'id_target': train_pairs.attrs['id_target']})
        transform.transform(predict_pairs, {'poserr': predict_pairs.attrs['poserr'], 
                                            'id_target': predict_pairs.attrs['id_target']})
        
        pipeline = Pipeline(model)
        pipeline.fit(train_pairs)

        data = pipeline.predict(predict_pairs, keep_features=False)
        data.query(f'P_0<{P_0_lim} and P_i>{P_i_lim} and flag_best', inplace=True)
        col = data.attrs['ids'] + [f"distance_to_{df_sub.attrs['name']}"]
        if one_to_one: 
            data = data.sort_values('P_i', ascending=False)
            data = data.drop_duplicates(subset=[data.attrs['id_frame']], keep='first').reset_index(drop=True)

        ids = data.attrs['ids']
        data.attrs = {}
        data.attrs['ids'] = ids
        data.rename(columns={'distance': f'distance_to_{df_sub.attrs["name"]}'}, inplace=True)
        return data[col]
    
    def _connect_non_trainable_model(self, df_main: pd.DataFrame, df_sub: pd.DataFrame, max_distance: float,
                        kwargs: dict = {}) -> pd.DataFrame:
        self._check_kwargs(kwargs, {'model': Model, 'one_to_one': bool, 'P_0_lim': float, 'P_i_lim': float})
        if kwargs['model'].fit_required:
            raise ValueError('Model does not require training. Use `trainable model` method instead.')
        model = kwargs['model']
        one_to_one = kwargs['one_to_one']
        P_0_lim = kwargs['P_0_lim']
        P_i_lim = kwargs['P_i_lim']

        predict_pairs = get_pairs_predict(df_main, df_sub, max_distance=max_distance)
        pipeline = Pipeline(model)

        data = pipeline.predict(predict_pairs, keep_features=False)
        data.query(f'P_0<{P_0_lim} and P_i>{P_i_lim} and flag_best', inplace=True)
        col = data.attrs['ids'] + [f"distance_to_{df_sub.attrs['name']}"] # , 'P_0', 'P_i', 'p_i'
        if one_to_one: 
            data = data.sort_values('P_i', ascending=False)
            data = data.drop_duplicates(subset=[data.attrs['id_frame']], keep='first').reset_index(drop=True)

        ids = data.attrs['ids']
        data.attrs = {}
        data.attrs['ids'] = ids
        data.rename(columns={'distance': f'distance_to_{df_sub.attrs["name"]}'}, inplace=True)
        return data[col]

    def get_linktable(self, method: str, df_main: pd.DataFrame, df_sub: pd.DataFrame, 
                    max_distance: float, kwargs: dict = {}) -> pd.DataFrame:
        self._check_attrs(df_main, df_sub)
        self._check_duplicates(df_main)
        self._check_duplicates(df_sub)
        if method == 'nearest':
            return self._connect_nearest(df_main, df_sub, max_distance, kwargs)
        if method == 'trainable model':
            return self._connect_trainable_model(df_main, df_sub, max_distance, kwargs)
        if method == 'non-trainable model':
            return self._connect_non_trainable_model(df_main, df_sub, max_distance, kwargs)
        
    def _join_df_in_linktable(self, df: pd.DataFrame, connection_df: pd.DataFrame, joined_id: str, outer_id: str) -> pd.DataFrame:
        """
        Join dataset of sources into connection table.

        TODO.
        """
        joined_inner = df.join(connection_df.set_index(joined_id), on=joined_id, how='inner')
        joined_outer = df[~df[joined_id].isin(joined_inner[joined_id])].copy()
        joined_outer[outer_id] = -1
        joined = pd.concat([joined_inner, joined_outer])
        joined.attrs = df.attrs
        return joined
        
    def _make_frame_table(self, linktables: List[pd.DataFrame], id_main: str) -> pd.DataFrame:
        """
        Create frame table with id_main and all other id in df in list.

        To make the frame, all connection table should contain one same column with id of sources (name of the columns should be the same).
        """
        all_main_ids = {item for table in linktables for item in table[id_main]}

        frame = pd.DataFrame({id_main: list(all_main_ids)})
        frame.attrs['ids'] = [id_main]
        frame.attrs['id'] = id_main
        for linktable in linktables:
            id_sub = linktable.attrs['ids'][0] if linktable.attrs['ids'][1] == id_main else linktable.attrs['ids'][1]
            frame = self._join_df_in_linktable(frame, linktable, id_main, id_sub)
            frame.attrs['ids'].append(id_sub)
        frame = move_columns_ahead(frame, frame.attrs['ids'])
        return frame
    
    def _join_catalogues_in_frame(self, frame: pd.DataFrame, dfs: List[pd.DataFrame]) -> pd.DataFrame:
        """
        Join all df in list in frame.

        Frame should contain id columns of all df.

        Args:
            frame: Frame with id columns.
            dfs: List of dataframes. Each df should contain id column, which is presented in frame.
        """
        frame.attrs['coords'] = [] 
        frame.attrs['mags'] = []
        frame.attrs['distances'] = []
        frame.attrs['poserrs'] = []
        for df in dfs:
            id_sub = df.attrs['id']
            old_columns = df.columns
            old_attrs = df.attrs.copy()
            add_postfix_to_all_columns_except(df, df.attrs['name'], get_attrs_columns(df, ['id', 'ra', 'dec', 'poserr']))
            frame = merge_save_attrs(frame, df, on=id_sub, how='left')

            frame.attrs['coords']+=get_attrs_columns(df, ['ra', 'dec'])
            if 'mags' in df.attrs:
                frame.attrs['mags']+=df.attrs['mags']
            frame.attrs['distances'].append(f'distance_to_{df.attrs["name"]}')
            frame.attrs['poserrs'].append(df.attrs['poserr'])

            df.columns = old_columns
            df.attrs = old_attrs

            if id_sub==frame.attrs['id']:
                frame.attrs['ra'], frame.attrs['dec'], frame.attrs['name'], frame.attrs['poserr'] =\
                      get_attrs_columns(df, ['ra', 'dec', 'name', 'poserr'])
                frame.attrs['name'] += '+connected'
                frame.attrs['distances'] = frame.attrs['distances'][:-1]

        return frame
    
    def connect_all_linktables(self, linktables: List[pd.DataFrame], dfs: List[pd.DataFrame], id_main: str) -> pd.DataFrame:
        """
        Join all linktables in frame.
        """
        frame = self._make_frame_table(linktables, id_main)
        return self._join_catalogues_in_frame(frame, dfs)

# Legacy
def connection_table_after_correlation(df_cor: pd.DataFrame, query: str = "P_0<0.05 and P_i>0.9 and flag_best==1",
                                       keep_distance: bool = True, make_unique: bool = False) -> pd.DataFrame:
    """
    Keep only best sources after smart-correlation.

    Args:
        df_cor: Correlated dataframe with columns ['P_i', 'P_0', 'flag_best'].
        query: Query for filtering.
        keep_distance: Keep distance between sources in df_main and df_sub in the output.
        make_unique: Make id_main unique. If after smart-correlation there are sources with several pairs with flag_best==1,\
            keep only one pair (it may be the result of duplicates in optical catalog).
    Returns:
        Correlated dataframe with columns id_main, id_sub and (optional) distance.
    """
    df = df_cor.query(query)
    col = df.attrs['ids']
    if keep_distance:
        col.append('distance')
    if make_unique:
        df = df.sort_values('P_i', ascending=False).drop_duplicates(subset=[df.attrs['id_frame']], keep='first').reset_index(drop=True)
    df.attrs['ids'] = df_cor.attrs['ids']
    return df[col].reset_index(drop=True)

# Legacy
def make_frame(connection_tables: List[pd.DataFrame], id_main: str) -> pd.DataFrame:
    """
    Create frame table with id_main and all other id in df in list.

    To make the frame, all connection table should contain one same column with id of sources (name of the columns should be the same).
    """
    #intersection = set(connection_tables[0].columns)
    #for i in connection_tables[1:]:
    #    intersection = intersection.intersection(i)
    #id_main = id_main

    ids = {item for table in connection_tables for item in table[id_main]}

    frame = pd.DataFrame({id_main: list(ids)})
    frame.attrs['ids'] = [id_main]
    frame.attrs['id'] = id_main
    for table in connection_tables:
        id_sub = table.attrs['ids'][0] if table.attrs['ids'][1] == id_main else table.attrs['ids'][1]
        frame = join_connection_table(frame, table, id_main, id_sub)
        frame.attrs['ids'].append(id_sub)
    return frame

# Legacy
def join_df_in_frame(frame: pd.DataFrame, dfs: List[pd.DataFrame]) -> pd.DataFrame:
    """
    Join all df in list in frame.

    Frame should contain id columns of all df.

    Args:
        frame: Frame with id columns.
        dfs: List of dataframes. Each df should contain id column, which is presented in frame.
    """
    frame.attrs['coords'] = [] 
    for df in dfs:
        id_sub = df.attrs['id']
        frame = merge_save_attrs(frame, df, on=id_sub, how='left')
        if id_sub==frame.attrs['id']:
            frame.attrs['ra'] = df.attrs['ra']
            frame.attrs['dec'] = df.attrs['dec']
            frame.attrs['name'] = df.attrs['name']+'+connected'
        frame.attrs['coords']+=[df.attrs['ra'], df.attrs['dec']]
    return frame

# Legacy
def mark_object_in_catalog(df_main: pd.DataFrame, catalog: pd.DataFrame, distance: float=0.2, hist=False) -> pd.DataFrame:
    """
    Mark rows with sources, which are represented in the second catalog (near the objects from this catalog).

    Args:
        df_main: Dataframe with sources, where is_in attribute will be added. Attrs: ['id', 'ra', 'dec', 'name']
        catalog: Catalog with sources, near which will be marked. Attrs: ['id', 'ra', 'dec', 'name']
        distance: Distance in degrees to find pairs.
    Returns:
        df_main with new column is_in_{catalog.attrs['name']}. Attrs: ['id', 'ra', 'dec', 'name'] + ['is_in']
    Note:
        df_main will be modified in place. New column is_in_{catalog.attrs['name']} will be added.
    """
    cor = correlate(df_main, catalog, distance, add_distance=True)
    cor['distance'].hist(bins=min(100, len(cor))) if hist else None
    df_main['is_in_'+catalog.attrs['name']] = df_main[df_main.attrs['id']].isin(cor[df_main.attrs['id']])

    df_main.attrs['is_in'] = [] if 'is_in' not in df_main.attrs else df_main.attrs['is_in'] + [catalog.attrs['name']]
    df_main.attrs['is_in'] = list(set(df_main.attrs['is_in']))

    return df_main





import pandas as pd
import numpy as np

from ..utils import add_postfix_to_main_columns, add_postfix_to_columns, add_magnitudes_to_attrs
from ..match.union import make_union_table, join_union_table, merge_columns_union_table

def read_catwise() -> pd.DataFrame:
    catwise = pd.read_csv('DS/4xmm_neighbours/CatWISE.csv')
    catwise = catwise[['Name', 'RAdeg', 'DEdeg', 'W1mproPM', 'W2mproPM', 'errHalfMaj']].drop_duplicates()
    catwise.rename(columns={'Name': 'id', 'RAdeg': 'ra', 'DEdeg': 'dec', 'errHalfMaj':'poserr',
                            'W1mproPM': 'w1_mag', 'W2mproPM': 'w2_mag'}, inplace=True)

    catwise.attrs['name'] = 'catwise'
    add_postfix_to_columns(catwise, 'catwise', ['id', 'ra', 'dec', 'poserr'], add_to_attrs=True)
    add_magnitudes_to_attrs(catwise, ['w1_mag', 'w2_mag'], 'catwise')
    return catwise

def read_allwise() -> pd.DataFrame:
    allwise = pd.read_csv('DS/4xmm_neighbours/ALLWISE.csv')
    allwise = allwise[['AllWISE', 'RAJ2000', 'DEJ2000', 'eeMaj', 'W1mag', 'W2mag', 'W3mag', 'W4mag', 'Jmag', 'Hmag', 'Kmag']].drop_duplicates()
    allwise.rename(columns={'AllWISE': 'id', 'RAJ2000': 'ra', 'DEJ2000': 'dec', 'eeMaj': 'poserr',
                            'W1mag': 'w1_mag', 'W2mag': 'w2_mag', 'W3mag': 'w3_mag', 'W4mag': 'w4_mag', 
                            'Jmag': 'j_mag', 'Hmag': 'h_mag', 'Kmag': 'k_mag'}, inplace=True)

    allwise.attrs['name'] = 'allwise'
    add_postfix_to_columns(allwise, 'allwise', ['id', 'ra', 'dec', 'poserr'], add_to_attrs=True)
    add_magnitudes_to_attrs(allwise, ['w1_mag', 'w2_mag', 'w3_mag', 'w4_mag', 'j_mag', 'h_mag', 'k_mag'], 'allwise')
    return allwise


def read_2mass() -> pd.DataFrame:
    mass = pd.read_csv('DS/4xmm_neighbours/2MASS.csv')
    mass = mass[['2MASS', 'RAJ2000', 'DEJ2000', 'errHalfMaj'] + ['Jmag', 'Hmag', 'Kmag']].drop_duplicates() # , 'e_Jmag', 'e_Hmag', 'e_Kmag'
    mass.rename(columns={'2MASS': 'id', 'RAJ2000': 'ra', 'DEJ2000': 'dec', 'errHalfMaj':'poserr',
                         'Jmag': 'j_mag', 'Hmag': 'h_mag', 'Kmag': 'k_mag'}, inplace=True)

    mass.attrs['name'] = 'mass'
    add_postfix_to_columns(mass, 'mass', ['id', 'ra', 'dec', 'poserr'], add_to_attrs=True)
    add_magnitudes_to_attrs(mass, ['j_mag', 'h_mag', 'k_mag'], 'mass')
    return mass

def read_glimpse() -> pd.DataFrame:
    glimpse = pd.read_csv('DS/4xmm_neighbours/GLIMPSE.csv')
    glimpse = glimpse[['GLIMPSE', 'RAdeg', 'DEdeg', 'errHalfMaj', '3.6mag', '4.5mag', '5.8mag', '8.0mag']].drop_duplicates()
    glimpse.rename(columns={'GLIMPSE': 'id', 'RAdeg': 'ra', 'DEdeg': 'dec', 'errHalfMaj':'poserr',
                            '3.6mag': '3.6_mag', '4.5mag': '4.5_mag', '5.8mag': '5.8_mag', '8.0mag': '8.0_mag'}, inplace=True)
    glimpse = glimpse.groupby(['id']).max().reset_index()

    glimpse.attrs['name'] = 'glimpse'
    add_postfix_to_columns(glimpse, 'glimpse', ['id', 'ra', 'dec', 'poserr'], add_to_attrs=True)
    add_magnitudes_to_attrs(glimpse, ['3.6_mag', '4.5_mag', '5.8_mag', '8.0_mag'], 'glimpse')
    return glimpse

def read_iphas() -> pd.DataFrame:
    iphas = pd.read_csv('DS/4xmm_neighbours/IPHAS.csv')
    iphas = iphas[['name', 'ra', 'dec', 'radec_err', 'r', 'i', 'ha']].drop_duplicates()
    iphas.rename(columns={'name': 'id', 'radec_err': 'poserr',
                          'r': 'r_mag', 'i': 'i_mag', 'ha': 'ha_mag'}, inplace=True)
    iphas = iphas.groupby(['id']).max().reset_index() # Надо придумать модуль для решения таких проблем

    iphas.attrs['name'] = 'iphas'
    add_postfix_to_columns(iphas, 'iphas', ['id', 'ra', 'dec', 'poserr'], add_to_attrs=True)
    add_magnitudes_to_attrs(iphas, ['r_mag', 'i_mag', 'ha_mag'], 'iphas')
    return iphas

def read_vphas() -> pd.DataFrame:
    vphas = pd.read_csv('DS/4xmm_neighbours/VPHAS.csv')
    vphas = vphas[['sourceID', 'RAJ2000', 'DEJ2000', 'rmag', 'imag', 'Hamag', 'umag', 'r2mag', 'gmag']].drop_duplicates()
    vphas.rename(columns={'sourceID': 'id', 'RAJ2000': 'ra', 'DEJ2000': 'dec',
                          'rmag': 'r_mag', 'imag': 'i_mag', 'Hamag': 'ha_mag', 'umag': 'u_mag', 'r2mag': 'r2_mag', 'gmag': 'g_mag'}, inplace=True)

    vphas.attrs['name'] = 'vphas'
    add_postfix_to_columns(vphas, 'vphas', ['id', 'ra', 'dec'], add_to_attrs=True)
    add_magnitudes_to_attrs(vphas, ['r_mag', 'i_mag', 'ha_mag', 'u_mag', 'r2_mag', 'g_mag'], 'vphas')
    return vphas

def read_ukidss() -> pd.DataFrame:
    ukidss = pd.read_csv('DS/4xmm_neighbours/UKIGPSdr11.csv')
    ukidss = ukidss[['ra_uki', 'dec_uki', 'J_AperMag3', 'H_AperMag3', 'K_AperMag3']].drop_duplicates()
    ukidss.rename(columns={'ra_uki': 'ra', 'dec_uki': 'dec',
                           'J_AperMag3': 'j_mag', 'H_AperMag3': 'h_mag', 'K_AperMag3': 'k_mag'}, inplace=True)
    ukidss['id'] = np.arange(len(ukidss))
    ukidss = ukidss[['id', 'ra', 'dec', 'j_mag', 'h_mag', 'k_mag']]

    for col in ['j_mag', 'h_mag', 'k_mag']:
        ukidss.loc[ukidss[col]<0, col] = np.NaN

    ukidss.attrs['name'] = 'ukidss'
    add_postfix_to_columns(ukidss, 'ukidss', ['id', 'ra', 'dec'], add_to_attrs=True)
    add_magnitudes_to_attrs(ukidss, ['j_mag', 'h_mag', 'k_mag'], 'ukidss')
    return ukidss

def make_ukidss_mass_union():
    ukidss = read_ukidss()
    mass = read_2mass()

    union_uki_mass = make_union_table(ukidss, mass).fillna(-1)
    union_table = join_union_table(union_uki_mass, ukidss, mass)
    union_table = merge_columns_union_table(union_table, {'j_mag': ('j_mag_ukidss', 'j_mag_mass'),
                                                        'h_mag': ('h_mag_ukidss', 'h_mag_mass'),
                                                        'k_mag': ('k_mag_ukidss', 'k_mag_mass')}, method='min', drop=True)
    del union_table['poserr_mass']
    return union_table


def read_xmm() -> pd.DataFrame:
    xmm = pd.read_pickle('DS/XMM/4XMM_clear_mag.pickle')
    xmm.attrs['name'] = 'xmm'
    add_postfix_to_columns(xmm, 'xmm', ['id', 'ra', 'dec', 'poserr'], add_to_attrs=True)
    add_magnitudes_to_attrs(xmm, ['mag_1', 'mag_2', 'mag_3', 'mag_4', 'mag_5', 'mag_8', 'mag_9'], 'xmm')
    
    return xmm

def read_csc() -> pd.DataFrame:
    csc = pd.read_pickle('DS/CSC/CSC_clear.pickle')
    csc = csc[['id', 'ra', 'dec']]
    csc.attrs['name'] = 'xmm'
    add_postfix_to_columns(csc, 'csc', ['id', 'ra', 'dec'], add_to_attrs=True)
    return csc

def read_gaia() -> pd.DataFrame:
    gaia = pd.read_pickle('DS/GAIA/GAIA_near_4XMM_15.pickle')
    gaia.attrs['name'] = 'gaia'
    add_postfix_to_columns(gaia, 'gaia', ['id', 'ra', 'dec'], add_to_attrs=True)
    add_magnitudes_to_attrs(gaia, ['g_mag', 'bp_mag', 'rp_mag'], 'gaia')
    return gaia

def read_ps() -> pd.DataFrame:
    ps = pd.read_pickle('DS/PS/PS_near_4XMM_15_mag.pickle')
    ps.attrs['name'] = 'ps'
    add_postfix_to_columns(ps, 'ps', ['id', 'ra', 'dec', 'poserr'], add_to_attrs=True)
    add_magnitudes_to_attrs(ps, ['rPSF_mag', 'iPSF_mag', 'zPSF_mag', 'yPSF_mag', 'gPSF_mag',
                                 'rKron_mag', 'iKron_mag', 'zKron_mag', 'yKron_mag', 'gKron_mag'], 'ps')
    return ps


from .duplicates import plot_duplicates_problem
from .duplicates import merge_duplicates_nearest
from .duplicates import merge_duplicates_same_id
from .duplicates import merge_duplicates_same_coords

from .gaps import fill_gaps
from .gaps import show_gaps

__all__ = ['plot_duplicates_problem', 
           'merge_duplicates_nearest',
           'merge_duplicates_same_id',
           'merge_duplicates_same_coords',
           
           'fill_gaps',
           'show_gaps']


import pandas as pd

def fill_gaps(df: pd.DataFrame, agg_functions: dict) -> pd.DataFrame:
    """
    Fills missing values in the DataFrame with known values.

    This function takes a DataFrame and a dictionary of aggregation methods to fill missing values in specific columns.
    The aggregation methods can be 'mean', 'max', 'min', 'avg', or a custom function that takes an array and returns a value.

    Args:
        df: The DataFrame in which missing values need to be filled.
        agg_functions: A dictionary where the keys are the aggregation methods and the values are lists of column names.

    Returns:
        The input DataFrame with missing values filled.

    Note:
        This function modifies the DataFrame in-place.

    Example:
        >>> df = pd.DataFrame({'A': [1, 2, None, 4], 'B': [5, None, 7, 8]})
        >>> fill_gaps(df, {'mean': ['A'], 'max': ['B']})
           A  B
        0  1  5
        1  2  7
        2  3  7
        3  4  8
    """
    for method in agg_functions:
        for col in agg_functions[method]: 
            value = df.loc[df[col].notna(), col].apply(method)
            df.loc[df[col].isna(), col] = value
    return df

def show_gaps(df: pd.DataFrame) -> pd.DataFrame:
    """Возвращает названия колонок с потоками в нужных фильтрах."""
    return df.isna().mean()




from sklearn.manifold import TSNE
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import roc_auc_score
from matplotlib import pyplot as plt
from typing import Union, Callable, Tuple, Any
import pandas as pd
import numpy as np
from collections import Counter

from ..utils import add_postfix_to_main_columns, correlate
from ..match.connection import mark_object_in_catalog


def make_data():
    data = pd.read_pickle('DS/find_hmxb/initial_data_mass.pickle')
    data = data.query('id_xmm!=-1 and abs(b_deg)<20')

    wise = pd.read_csv('DS/tmp/all_cat.csv')[['main_id', 'ra', 'dec','W1mag', 'W2mag', 'W3mag', 'W4mag', 'main_type']]
    wise.columns = ['id_wise', 'ra_wise', 'dec_wise', 'W1mag', 'W2mag', 'W3mag', 'W4mag', 'otype']

    wise = pd.read_csv('DS/tmp/all_cat.csv')[['main_id', 'ra', 'dec','W1mag', 'W2mag', 'W3mag', 'W4mag', 'main_type']]
    wise.columns = ['id_wise', 'ra_wise', 'dec_wise', 'W1mag', 'W2mag', 'W3mag', 'W4mag', 'otype']

    for i in wise.columns:
        data[i] = wise[i].values

    if True:
        data['otype'] = data['otype'].replace({'CV*': 'CV', 
                                            'G': 'Galaxy', 
                                            'QSO': 'Quasar',
                                            
                                            'BLL': 'Quasar',
                                            "Sy1": 'Galaxy',
                                            'Sy2': 'Galaxy', 
                                            'GiC': 'Galaxy',
                                            'GiG': 'Galaxy',
                                            'AGN': 'Galaxy',
                                            'rG': 'Galaxy'
                                            })

    data.attrs['ids'].append('id_wise')
    data.attrs['coord'].append('dec_wise')
    data.attrs['coord'].append('ra_wise')
    data.attrs['names'].append('wise')

    data['w1-w2'] = data['W1mag'] - data['W2mag']
    data['new_parallax'] = data['parallax'].values/ data['parallax_error'].values

    del data['parallax'], data['parallax_error']

    return data



def tsne(data: pd.DataFrame, ignored: list, verbose: bool = False, random_state: int = 42,
         perplexity: int = 400, scalar: bool = True, n_components: int = 2) -> pd.DataFrame:

    model = TSNE(n_components=n_components, learning_rate='auto', init='random', perplexity=perplexity,
                 random_state=random_state, verbose=verbose)

    data = data.drop(columns=ignored)

    for i in data.columns:
        data[i] = data[i].fillna(data.loc[data[i].notna(), i].quantile(0.95))
        if data[i].isna().sum() > 0:
            del data[i]

    if scalar:
        data = StandardScaler().fit_transform(data)
    X_tr = model.fit_transform(data)
    return X_tr


def plot_hmxb(compressed: np.ndarray, train: pd.DataFrame, ax: Union[plt.Axes, None] = None, name: str = '') -> None:  # *(train.g_mag<12)

    if ax is None:
        fig, ax = plt.subplots()
    compressed_false = compressed[train.type=='None']
    ax.scatter(compressed_false[:, 0], compressed_false[:, 1], s=0.1, c='g', label='Unknown')

    color = {
        #'CV': 'orange', 'Quasar': 'black', 'Star': 'purple',
        'LMXB': 'b',  
        'HMXB': 'r', }

    for key, val in color.items():
        x = compressed[train.type.values == key]
        ax.scatter(x[:, 0], x[:, 1], s=6, c=val, label=key)

    plt.legend()


    if name:
        ax.set_title(name)

def plot_hmxb(compressed: np.ndarray, train: pd.DataFrame, ax: Union[plt.Axes, None] = None, name: str = '',scaler: int=1) -> None:  # *(train.g_mag<12)

    if ax is None:
        fig, ax = plt.subplots()
    compressed_false = compressed[train['otype'].isna()]
    ax.scatter(compressed_false[:, 0], compressed_false[:, 1], s=0.05, c='g', label='Unknown')

    if 'otype' in train.columns:
        color = {
        #'Star': 'purple', 
        #'XRAY': 'orange', 'YSO': 'black',
        #'LMXB': 'b',  
        'HMXB': 'r', 
        #'HMXB_CAND': 'b'
        }
        for key, val in color.items():
            if key=='HMXB' or key=='HMXB_CAND':
                x = compressed[(train['otype'].values == key)| (train['type']==key)]
                ax.scatter(x[:, 0], x[:, 1], s=15*scaler, c=val, label=key)
                continue
            #print(key)
            x = compressed[(train.otype.values == key)| (train['type']==key)]
            #print(x.shape)
            ax.scatter(x[:, 0], x[:, 1], s=3*scaler, c=val, label=key)

    #plt.legend(fontsize='x-small', bbox_to_anchor=(1.1, 1.1))
    ax.set_xticks([])
    ax.set_yticks([])
    if name:
        ax.set_title(name + f' | HMXB:{train["type"].value_counts()["HMXB"]}, TOTAL:{train.shape[0]}')

def plot_hmxb_classes(compressed: np.ndarray, train: pd.DataFrame, ax: Union[plt.Axes, None] = None, name: str = '',scaler: int=1) -> None:  # *(train.g_mag<12)

    if ax is None:
        fig, ax = plt.subplots()
    compressed_false = compressed[train.otype.isna()]
    ax.scatter(compressed_false[:, 0], compressed_false[:, 1], s=0.001, c='b', label='Unknown')

    colors = ['red', 'blue', 'green', 'purple', 'yellow', 'pink']
    replace = {0:'red', 1:'blue', 2:'green', 3:'purple', 4:'yellow', 5:'pink'}
    compressed_hmxb = compressed[train['type']=='HMXB']
    ax.scatter(compressed_hmxb[:, 0], compressed_hmxb[:, 1], s=15*scaler, 
                c=train[train['type']=='HMXB']['last_claster'].replace(replace), label='HMXB')

    plt.legend(fontsize='x-small', bbox_to_anchor=(1.1, 1.1))
    ax.set_xticks([])
    ax.set_yticks([])
    if name:
        ax.set_title(name, fontsize='x-small')

def plot_hmxb_partial(compressed: np.ndarray, train: pd.DataFrame, ax: Union[plt.Axes, None] = None, name: str = '', classes: list = []) -> None:

    if ax is None:
        fig, ax = plt.subplots()
    compressed_false = compressed[train.type=='None']
    ax.scatter(compressed_false[:, 0], compressed_false[:, 1], s=0.001, c='b', label='Unknown')

    colors = ['black', 'orange', 'purple', 'g', 'magenta', 'pink', 'yellow' ]

    for num, color in enumerate(classes):
        x = compressed[(train.otype.values == classes[num])|(train.type.values == classes[num])]
        ax.scatter(x[:, 0], x[:, 1], s=20, c=colors[num], label=classes[num])

    ax.legend(loc='upper right',fontsize='small')


    if name:
        ax.set_title(name)

def plot_mag_shift(real_data, nn_data, ind=639, ax=None):
    import seaborn as sns
    maggas = ['g_mag', 'bp_mag', 'rp_mag', 'Jmag', 'Hmag', 'Kmag', 'rmag', 'imag', 'Hamag', 'mag_flux_1', 'mag_flux_2', 'mag_flux_3']
    stat = []
    for i in maggas:
        stat.append([real_data.iloc[ind][i], nn_data.iloc[ind][i], real_data[i].quantile(0.1), real_data[i].quantile(0.9)])

    stat = np.array(stat)

    maggas = ['g_mag', 'bp_mag', 'rp_mag', 'Jmag', 'Hmag', 'Kmag', 'rmag', 'imag', 'Hamag', 'Xmag_02-05', 'Xmag_05-1', 'Xmag_1-2']

    if ax is None:
        fig, ax = plt.subplots()

    ax.plot(maggas, stat[:,0], 'g-', label='real')
    ax.plot(maggas, stat[:,1], 'b-', label='nn')

    # Add boxplot to each point
    for x, (x_val, y_val) in enumerate(zip(maggas, stat)):
        sns.boxplot(x=[x_val, x_val], y=[stat[x, 2], stat[x, 3]], ax=ax, width=0.1, color='red')

    ax.tick_params(axis='x', labelrotation=90)


# ----------------------------------------------------------------------------------------------- #
#                                         Gat rank                                                #
# ----------------------------------------------------------------------------------------------- #


def dist_to_n_nearest_(n_coords: np.ndarray, m_coords: np.ndarray, count_nearest: int = 5, how: str = 'mean') -> np.ndarray:
    """
    ['mean', 'exp', 'harmonic', 'geometric']
    """
    nearest_distances = []
    for m_point in m_coords:
        distances = np.sqrt(np.sum((n_coords - m_point)**2, axis=1))
        nearest_indices = np.argsort(distances)[:count_nearest]
        if how == 'mean':
            mean = distances[nearest_indices].mean()
        elif how == 'exp':
            mean = np.average(distances[nearest_indices], weights=np.exp(-0.5 * np.arange(len(nearest_indices))))
        elif how == 'harmonic':
            mean = np.average(distances[nearest_indices], weights=1 / (np.arange(len(nearest_indices)) + 1))
        elif how == 'geometric':
            mean = np.average(distances[nearest_indices], weights=1 / np.log(np.arange(len(nearest_indices)) + 2))
        nearest_distances.append(mean)
    return np.array(nearest_distances)

from scipy.spatial import cKDTree

def dist_to_n_nearest(n_coords: np.ndarray, m_coords: np.ndarray, count_nearest: int = 5, how: str = 'mean') -> np.ndarray:
    """
    ['mean', 'exp', 'harmonic', 'geometric']
    """
    ntree = cKDTree(n_coords)
    nearest_distances = []

    for m_point in m_coords:
        distances, _ = ntree.query(m_point, k=count_nearest)
        
        if how == 'mean':
            mean = np.mean(distances)
        elif how == 'exp':
            mean = np.average(distances, weights=np.exp(-0.5 * np.arange(len(distances))))
        elif how == 'harmonic':
            mean = np.average(distances, weights=1 / (np.arange(len(distances)) + 1))
        elif how == 'geometric':
            mean = np.average(distances, weights=1 / np.log(np.arange(len(distances)) + 2))
        
        nearest_distances.append(mean)
    
    return np.array(nearest_distances)


# ----------------------------------------------------------------------------------------------- #
#                                  метрика отбора метода                                          #
# ----------------------------------------------------------------------------------------------- #
def _number_top_n(ranks: np.ndarray, n: int) -> int:
    return (ranks <= n).sum()


def number_top_n(sorted_data: pd.DataFrame, mark: np.ndarray, n: int) -> int:
    return _number_top_n(sorted_data[mark].index, n)


def number_top_10(sorted_data: pd.DataFrame, mark: np.ndarray) -> int:
    return _number_top_n(sorted_data[mark].index, 10)


def number_top_100(sorted_data: pd.DataFrame, mark: np.ndarray) -> int:
    return _number_top_n(sorted_data[mark].index, 100)


def roc_auc_rank(sorted_data: pd.DataFrame, mark: np.ndarray) -> float:
    return roc_auc_score(mark, -sorted_data.index)


def evaluate_compression(compress: np.ndarray, data: pd.DataFrame, samples: int = 10, sample_size: int = 20,
                         metric: str = 'mean', n_nearest: int = 5,
                         score_fn: Callable = lambda x, y: number_top_n(x, y, n=10)) -> float:
    scores = []
    for i in range(samples):
        data_ = data[['id_gaia', 'type', 'otype']].copy()
        #data_.loc[data_['otype']=='HMXB', 'type'] = 'HMXB'
        data_['id_gaia'] = data_['id_gaia'].astype(np.int64)
        id_hmxb = data_.query('type=="HMXB"')['id_gaia'].sample(sample_size).values

        data_['metric'] = dist_to_n_nearest(compress[(data_['type']=='HMXB') * (~data_.id_gaia.isin(id_hmxb))],
                                            compress, n_nearest, metric)
        data_.sort_values('metric', inplace=True)
        data_ = data_[(data_['type'] != 'HMXB') | (data_['id_gaia'].isin(id_hmxb))].reset_index(drop=True)
        score = score_fn(data_, data_['id_gaia'].isin(id_hmxb))
        scores.append(score)
    return np.array(scores).mean()


# ----------------------------------------------------------------------------------------------- #
#                                  подготовка кандидатов                                          #
# ----------------------------------------------------------------------------------------------- #

def get_metric(data: pd.DataFrame, compressed: np.ndarray, how: str = 'mean',
               n_neighbour: int = 10, metric_to_all: bool = False) -> np.ndarray:
    if metric_to_all:
        X_tr_false = compressed
    else:
        X_tr_false = compressed[(~data.is_hmxb.values.astype(bool))]
    X_tr_gaia = compressed[data.is_hmxb.values.astype(bool)]
    metric = dist_to_n_nearest(X_tr_gaia, X_tr_false, n_neighbour, how)
    return metric


def prepare_candidates(train: pd.DataFrame, metric: np.ndarray) -> pd.DataFrame:
    candidates = train[['id_gaia', 'id_xmm', 'id_ps', 'ra_gaia', 'dec_gaia', 'ra_xmm',
                        'dec_xmm', 'g_mag', 'p_quasar', 'p_galaxy', 'p_star', 'is_lmxb']].loc[~train.is_hmxb.values.astype(bool)]

    candidates['metric'] = metric
    candidates = candidates.sort_values('metric')
    candidates.reset_index(inplace=True, drop=True)
    candidates = candidates.rename(columns={'metric': 'score'})
    candidates['p_quasar'] = candidates['p_quasar'].round(2)
    candidates['p_galaxy'] = candidates['p_galaxy'].round(2)
    candidates['p_star'] = candidates['p_star'].round(2)
    candidates['score'] = candidates['score'].round(2)
    candidates['to_copy'] = candidates['ra_xmm'].astype(str) + ' ' + candidates['dec_xmm'].astype(str)
    candidates['url_xray'] = f'http://cdsportal.u-strasbg.fr/?target=' + candidates["ra_xmm"].astype(str) + '%20' +\
        candidates["dec_xmm"].astype(str)
    candidates['url_gaia'] = f'http://cdsportal.u-strasbg.fr/?target=' + candidates["ra_gaia"].astype(str) + '%20' +\
        candidates["dec_gaia"].astype(str)

    xmm_ = pd.read_pickle('DS/XMM/4XMM_raw.pickle')
    xmm_ = xmm_[['srcid', 'sc_hr1', 'sc_hr2', 'sc_hr3', 'sc_hr4', 'sc_ep_2_flux', 'sc_ep_3_flux']].rename(columns={'srcid': 'id_xmm', 'sc_hr1': 'hr1',
                                                                                                                   'sc_hr2': 'hr2', 'sc_hr3': 'hr3', 'sc_hr4': 'hr4', })
    xmm_['flux_0.5-2'] = xmm_['sc_ep_2_flux'] + xmm_['sc_ep_3_flux']
    del xmm_['sc_ep_2_flux'], xmm_['sc_ep_3_flux']
    candidates = pd.merge(candidates, xmm_, on='id_xmm', how='left')

    old = pd.read_csv('DS/find_hmxb/old_candidates.csv')
    old = old.head(50)
    candidates['was_last_top50'] = candidates.id_xmm.isin(old.id_xmm)

    candidates = candidates[['id_gaia', 'id_xmm', 'id_ps', 'ra_gaia', 'dec_gaia', 'g_mag', 'ra_xmm', 'dec_xmm',
                            'flux_0.5-2', 'hr1', 'hr2', 'hr3', 'hr4', 'p_quasar', 'p_galaxy', 'p_star',
                             'score', 'was_last_top50', 'is_lmxb', 'to_copy', 'url_xray', 'url_gaia']]

    candidates['id_gaia'] = candidates['id_gaia'].astype(str)
    candidates['id_xmm'] = candidates['id_xmm'].astype(str)
    candidates['id_ps'] = candidates['id_ps'].astype(str)
    return candidates

    
# ----------------------------------------------------------------------------------------------- #
#                                          Добавить классы                                        #
# ----------------------------------------------------------------------------------------------- #


def make_star1():
    star_1 = pd.read_pickle('DS/find_hmxb/mark_in_catalog/star_1.pkl')
    star_1 = star_1[['recno', '_RA', '_DE']].rename(columns={'recno':'id', '_RA': 'ra', '_DE':'dec'})
    add_postfix_to_main_columns(star_1, 'star1')
    star_1.attrs['name'] = 'star1'
    star_1.attrs['id'] = 'id_star1'
    star_1.attrs['ra'] = 'ra_star1'
    star_1.attrs['dec'] = 'dec_star1'

    return star_1

def make_star2():
    star_2 = pd.read_pickle('DS/find_hmxb/mark_in_catalog/star_2.pkl')
    star_2 = star_2[['recno', 'RAJ2000', 'DEJ2000']].rename(columns={'recno':'id', 'RAJ2000': 'ra', 'DEJ2000':'dec'}).loc[1:]
    add_postfix_to_main_columns(star_2, 'star2')
    star_2.attrs['name'] = 'star2'
    star_2.attrs['id'] = 'id_star2'
    star_2.attrs['ra'] = 'ra_star2'
    star_2.attrs['dec'] = 'dec_star2'

    star_2['ra_star2'] = star_2['ra_star2'].map(lambda x: float(x.split(' ')[0])*15 + float(x.split(' ')[1])*0.25 + float(x.split(' ')[2])*(0.25 / 60))
    star_2['dec_star2'] = star_2['dec_star2'].map(lambda x: float(x.split(' ')[0]) + float(x.split(' ')[1])*(1/60) + float(x.split(' ')[2])*(1/3600))
    return star_2

def make_quas():
    quas = pd.read_pickle('DS/find_hmxb/mark_in_catalog/mil_quasar.pkl')
    quas = quas[['recno', 'RAJ2000', 'DEJ2000']].rename(columns={'recno':'id', 'RAJ2000': 'ra', 'DEJ2000':'dec'}).copy()
    add_postfix_to_main_columns(quas, 'quas')
    quas.attrs['name'] = 'quas'
    quas.attrs['id'] = 'id_quas'
    quas.attrs['ra'] = 'ra_quas'
    quas.attrs['dec'] = 'dec_quas'

    return quas

def make_cv_cat1():
    cv_cat1 = pd.read_pickle('DS/find_hmxb/mark_in_catalog/cv_cat1.pkl')
    cv_cat1 = cv_cat1[['recno', 'RAJ2000', 'DEJ2000']].rename(columns={'recno':'id', 'RAJ2000': 'ra', 'DEJ2000':'dec'}).copy()
    add_postfix_to_main_columns(cv_cat1, 'cv1')
    cv_cat1.attrs['name'] = 'cv1'
    cv_cat1.attrs['id'] = 'id_cv1'
    cv_cat1.attrs['ra'] = 'ra_cv1'
    cv_cat1.attrs['dec'] = 'dec_cv1'

    cv_cat1['ra_cv1'] = cv_cat1['ra_cv1'].map(lambda x: float(x.split(' ')[0])*15 + float(x.split(' ')[1])*0.25 + float(x.split(' ')[2])*(0.25 / 60))
    cv_cat1['dec_cv1'] = cv_cat1['dec_cv1'].map(lambda x: float(x.split(' ')[0]) + float(x.split(' ')[1])*(1/60) + float(x.split(' ')[2])*(1/3600))
    return cv_cat1

def make_cv_cat2():
    cv_cat2 = pd.read_pickle('DS/find_hmxb/mark_in_catalog/cv_cat2.pkl')
    cv_cat2 = cv_cat2[['recno', 'RAJ2000', 'DEJ2000']].rename(columns={'recno':'id', 'RAJ2000': 'ra', 'DEJ2000':'dec'}).copy()
    add_postfix_to_main_columns(cv_cat2, 'cv2')
    cv_cat2.attrs['name'] = 'cv2'
    cv_cat2.attrs['id'] = 'id_cv2'
    cv_cat2.attrs['ra'] = 'ra_cv2'
    cv_cat2.attrs['dec'] = 'dec_cv2'

    cv_cat2['ra_cv2'] = cv_cat2['ra_cv2'].map(lambda x: float(x.split(' ')[0])*15 + float(x.split(' ')[1])*0.25 + float(x.split(' ')[2])*(0.25 / 60))
    cv_cat2['dec_cv2'] = cv_cat2['dec_cv2'].map(lambda x: float(x.split(' ')[0]) + float(x.split(' ')[1])*(1/60) + float(x.split(' ')[2])*(1/3600))
    return cv_cat2

def make_lis():
    desi_lis = pd.read_pickle('DS/LS/LS_class.pickle')
    add_postfix_to_main_columns(desi_lis, 'lis')
    desi_lis.attrs['name'] = 'lis'
    desi_lis.attrs['id'] = 'id_lis'
    desi_lis.attrs['ra'] = 'ra_lis'
    desi_lis.attrs['dec'] = 'dec_lis'

    return desi_lis

def add_hmxb_and_lmxb(df_main):
    hmxb = pd.read_pickle('DS/HMXB/HMXB_2023.pickle')
    lmxb = pd.read_pickle('DS/HMXB/LMXB_2023.pickle')
    df_main['is_hmxb'] = df_main['id_gaia'].isin(hmxb['Gaia_DR3_ID']).astype(int)
    df_main['is_lmxb'] = df_main['id_gaia'].isin(lmxb['Gaia_DR3_ID']).astype(int)

    return df_main

def add_classes_real(df_main):
    quas = make_quas()
    star1 = make_star1()
    star2 = make_star2()
    cv1 = make_cv_cat1()
    cv2 = make_cv_cat2()
    lis = make_lis()

    df_main = mark_object_in_catalog(df_main, quas, 0.2)
    df_main = mark_object_in_catalog(df_main, cv1, 3)
    df_main = mark_object_in_catalog(df_main, cv2, 3)
    df_main = mark_object_in_catalog(df_main, star1, 5)
    df_main = mark_object_in_catalog(df_main, star2, 1)

    cor = correlate(df_main, lis, 0.5, add_distance=True)

    df_main = add_hmxb_and_lmxb(df_main)

    df_main['type'] = 'None'
    df_main.loc[df_main.is_in_quas==True, 'type'] = 'Quasar'
    df_main.loc[df_main.is_in_cv1==True, 'type'] = 'CV'
    df_main.loc[df_main.is_in_cv2==True, 'type'] = 'CV'
    df_main.loc[df_main.is_in_star1==True, 'type'] = 'Star'
    df_main.loc[df_main.is_in_star2==True, 'type'] = 'Star'
    df_main.loc[df_main.id_gaia.isin(cor.query('spectype==0')['id_gaia']), 'type'] = 'Quasar'
    df_main.loc[df_main.id_gaia.isin(cor.query('spectype==1')['id_gaia']), 'type'] = 'Galaxy'
    df_main.loc[df_main.id_gaia.isin(cor.query('spectype==2')['id_gaia']), 'type'] = 'Star'
    df_main.loc[df_main.is_hmxb==True, 'type'] = 'HMXB'
    df_main.loc[df_main.is_lmxb==True, 'type'] = 'LMXB'

    del df_main['is_in_quas'], df_main['is_in_cv1'], df_main['is_in_cv2'], df_main['is_in_star1'],\
          df_main['is_in_star2'], df_main['is_hmxb'], df_main['is_lmxb']

    return df_main


# ----------------------------------------------------------------------------------------------- #
#                                          Clastering                                             #
# ----------------------------------------------------------------------------------------------- #

def plot_classes(compressed, data,  classes, ax=None, name=''):
    if ax is None:
        fig, ax = plt.subplots()
    X = compressed[data.type=='HMXB']
    color = ['red', 'blue', 'green', 'purple', 'yellow', 'pink']
    ax.scatter(compressed[:,0], compressed[:,1], s=0.001)
    ax.scatter(X[:,0], X[:,1], color=[color[i] for i in classes], s=100)
    ax.set_title(name)

def prepare_classes(compressed, data, model, ax=None, name='', plot=True):
    X = compressed[data.type=='HMXB']
    classes = model.fit_predict(X)
    #while classes[0]!=0 or classes[6]!=2 or classes[8]!=3:
    #    classes = model.fit_predict(X)
    classes = np.array([i if i!=-1 else np.max(classes)+1 for i in classes])
    if plot:
        plot_classes(compressed, data, classes, ax, name)
    return classes

def plot_class_distribution(classes, data, hmxb, ax=None, n_classes=5, name=''):
    if ax is None:
        fig, ax = plt.subplots()

    class_to_id_gaia = lambda x: data.query('type=="HMXB"').loc[classes==x].id_gaia

    category_counts = Counter()
    for i in range(n_classes):
        id_gaia = class_to_id_gaia(i)
        hmxb_classes = ','.join(map(str, list(hmxb.query("id_gaia in @id_gaia").Xray_Type.values))).split(',')
        category_counts.update(hmxb_classes)
    
    sorted_categories = [cat for cat, _ in category_counts.most_common()]

    # Построение гистограммы с отсортированными категориями
    color = ['red', 'blue', 'green', 'purple', 'yellow', 'pink']  # замените на ваши цвета
    bottom = np.zeros(len(sorted_categories))

    for i in range(n_classes):  # замените на ваш реальный диапазон классов
        id_gaia = class_to_id_gaia(i)
        hmxb_classes = ','.join(map(str, list(hmxb.query("id_gaia in @id_gaia").Xray_Type.values))).split(',')
        counts = dict(Counter(hmxb_classes))
        values = [counts.get(cat, 0) for cat in sorted_categories]

        ax.bar(sorted_categories, values, bottom=bottom, color=color[i])
        bottom += np.array(values)
    ax.set_title(name)

    
def plot_class_distribution_separately(classes, data, hmxb, n_classes=5, names=['','','','',''], axes=None):
    if axes is None:
        fig, axes = plt.subplots(1, n_classes, figsize=(20, 4))

    class_to_id_gaia = lambda x: data.query('type=="HMXB"').loc[classes==x].id_gaia

    category_counts_all = Counter()
    for i in range(n_classes):
        id_gaia = class_to_id_gaia(i)
        hmxb_classes = ','.join(map(str, list(hmxb.query("id_gaia in @id_gaia").Xray_Type.values))).split(',')
        category_counts_all.update(hmxb_classes)

    sorted_categories = list(category_counts_all.keys())

    for i, ax in enumerate(axes):
        category_counts = Counter()
        id_gaia = class_to_id_gaia(i)
        hmxb_classes = ','.join(map(str, list(hmxb.query("id_gaia in @id_gaia").Xray_Type.values))).split(',')
        category_counts.update(hmxb_classes)

        color = ['red', 'blue', 'green', 'purple', 'yellow', 'pink']  # замените на ваши цвета
        bottom = np.zeros(len(sorted_categories))

        values = [category_counts.get(cat, 0) for cat in sorted_categories]

        ax.bar(sorted_categories, values, bottom=bottom, color=color[i])
        ax.set_title(f'Class {i}')
        ax.set_xlabel('HMXB Type')
        ax.set_ylabel('Count')
        ax.set_xticklabels(list(sorted_categories), rotation=90)
        print(i)
        ax.set_title(names[i])


def calc_centers(classes, compressed, data):
    coords = compressed[data['type']=='HMXB']
    centers = {}
    for class_ in set(classes):
        centers[class_] = np.mean(coords[classes==class_], axis=0)
    return centers

def plot_centers(centers, ax=None, name=''):
    if ax is None:
        fig, ax = plt.subplots()
    for i in centers:
        ax.scatter(centers[i][0], centers[i][1], color='black', s=150)
    ax.set_title(name)  


from ..match.connection import connection_table_nearest, connection_table_after_correlation
from ..utils import merge_save_attrs, add_postfix_to_main_columns, correlate, keep_nearest_pairs

from ..match import get_pairs_train, get_pairs_predict, Pipeline
from ..transforms import CoordinatesGalacticTransform, FullAstrometryTransform
from ..match.models import Catboost

from .find_hmxb import tsne
from .read_catalogues import read_xmm

import pandas as pd
import numpy as np

def add_otype(data):
    # ТУТ ДУБЛИКАТ, надо убрать
    df_typed = pd.read_csv('DS/new_hmxb/all_cat_with_simbad.csv')[['id_gaia', 'ra_gaia', 'dec_gaia', 'otype']]
    df_typed.rename(columns={'id_gaia': '_id_', 'ra_gaia': '_ra_', 'dec_gaia': '_dec_'}, inplace=True)
    df_typed.attrs['id'] = '_id_'; df_typed.attrs['ra'] = '_ra_'; df_typed.attrs['dec'] = '_dec_'; df_typed.attrs['name'] = 'gaia'

    connection = connection_table_nearest(data, df_typed[df_typed['otype'].notna()], max_dist=1, keep_distance=True)
    connection = keep_nearest_pairs(connection, '_id_', )
    connection = keep_nearest_pairs(connection, data.attrs['id'], )
    print(connection['_id_'].is_unique, connection[data.attrs['id']].is_unique)
    del connection['distance']
    data = merge_save_attrs(data, connection, on=data.attrs['id'], how='left')
    data = merge_save_attrs(data, df_typed[['_id_', 'otype']], on='_id_', how='left')
    print(data.shape)
    del data['_id_']
    data = data.drop_duplicates()
    return data

def add_type(data):
    hmxb = pd.read_pickle('DS/HMXB/HMXB_2023.pickle')
    data['type'] = np.NaN
    data.loc[data['id_xmm'].isin(hmxb['id_xmm']), 'type'] = 'HMXB'

    lmxb = pd.read_pickle('DS/HMXB/LMXB_2023.pickle')[['Name', 'RAdeg', 'DEdeg']]
    lmxb.rename(columns={'Name': 'id', 'RAdeg': 'ra', 'DEdeg': 'dec'}, inplace=True)
    add_postfix_to_main_columns(lmxb, 'lmxb')
    #print(lmxb.columns)
    cor = correlate(data, lmxb, 0.5, add_distance=False, add_attrs=False)
    id_ = data.attrs['id']
    data.loc[data[id_].isin(cor[id_]), 'type'] = 'LMXB'
    return data

def add_last_claster(data):
    tmp = pd.read_csv('DS/find_hmxb/baseline+wise.csv')
    tmp = tmp[tmp['claster'].notna()][['id_xmm', 'claster']]
    tmp.rename(columns={'claster': 'last_claster'}, inplace=True)
    data = merge_save_attrs(data, tmp, on='id_xmm', how='left')
    return data

def get_connection(opt, target, target_support, add_all_hmxb=True):
    train_pairs = get_pairs_train(opt, target, target_support, max_distance=5)
    FullAstrometryTransform((train_pairs.attrs['ra_frame'], train_pairs.attrs['dec_frame']), 
                             add_astro_factor=False).transform(train_pairs, {'poserr': train_pairs.attrs['poserr'], 
                                                                             'id_target': train_pairs.attrs['id_target']})

    pred_pairs = get_pairs_predict(opt, target, max_distance=5)
    FullAstrometryTransform((pred_pairs.attrs['ra_frame'], pred_pairs.attrs['dec_frame']), 
                             add_astro_factor=False).transform(pred_pairs, {'poserr': pred_pairs.attrs['poserr'], 
                                                                            'id_target': pred_pairs.attrs['id_target']})
    pipeline = Pipeline(Catboost())
    pipeline.fit(train_pairs)
    connection = pipeline.predict(pred_pairs, keep_features=False)
    if add_all_hmxb:
        connection_hmxb = connection[connection['id_xmm'].isin(pd.read_pickle('DS/HMXB/HMXB_2023.pickle')['id_xmm'])]
    #connection_hmxb.query('flag_best==1').to_pickle('DS/new_hmxb/tmp.pickle')
        connection_hmxb = connection_hmxb.query('flag_best==1 and P_0<0.5')[['id_xmm', pred_pairs.attrs['id_frame']]]
        connection = connection_table_after_correlation(connection, keep_distance=False, make_unique=True, query='flag_best==1 and P_i>0.8 and P_0<0.5')
        return pd.concat([connection_hmxb, connection]).drop_duplicates()
    else:
        connection = connection_table_after_correlation(connection, keep_distance=False, make_unique=True, query='flag_best==1 and P_i>0.8 and P_0<0.5')
        return connection

def add_classes(data):
    CoordinatesGalacticTransform((data.attrs['ra'], data.attrs['dec'])).transform(data)
    #print(data.shape)
    data = add_otype(data)
    #print(data.shape)
    data = add_type(data)
    #print(data.shape)
    data = add_last_claster(data)
    #print(data.shape)
    return data

def get_several_tsne(data, ignored, perplexities, random_state=42):
    tsne_results = {}
    for perplexity in perplexities:
        tsne_result = tsne(data, ignored, perplexity=perplexity, random_state=random_state)
        tsne_results[perplexity] = tsne_result
        print(perplexity)
    return tsne_results


def add_director():
    xmm = read_xmm()
    tmp = pd.read_csv('DS/new_hmxb/2obj.csv')
    tmp = tmp[['Source', 'ra_cone', 'dec_cone', 'b', 'l', 'jAperMag3', 'hAperMag3', 'ksAperMag3']]
    director = xmm[xmm['id_xmm'].isin(tmp['Source'])]
    director = pd.merge(director, tmp, how='left', left_on='id_xmm', right_on='Source')
    del director['Source']
    director['id_ukidss'] = director['id_xmm']

    director.rename(columns={'ra_cone': 'ra_ukidss', 'dec_cone': 'dec_ukidss',
                            'b': 'b_deg', 'l': 'l_deg', 'jAperMag3': 'j_mag_ukidss', 
                            'hAperMag3': 'h_mag_ukidss', 'ksAperMag3': 'k_mag_ukidss'}, inplace=True)
    director['type'] = 'HMXB_CAND'
    director['otype'] = np.nan
    director['last_claster'] = np.nan
    arrts = data.attrs
    data = pd.concat([data, director])
    data.attrs = arrts

def modify_otype(data):
    data['otype'] = data['otype'].replace({'CV*': 'CV', 
                                       'G': 'Galaxy', 
                                       'QSO': 'Quasar',
                                       
                                       'BLL': 'Quasar',
                                       "Sy1": 'Galaxy',
                                       'Sy2': 'Galaxy', 
                                       'GiC': 'Galaxy',
                                       'GiG': 'Galaxy',
                                       'AGN': 'Galaxy',
                                       'rG': 'Galaxy',

                                       'YSO_Candidate': 'YSO',
                                       'TTau*': 'XRAY',
                                       'SB*': 'XRAY'
                                     })
    hmxb_cand = pd.read_pickle('DS/new_hmxb/3HMXBs.pickle')
    data.loc[data['id_xmm'].isin(hmxb_cand['id']), 'type'] = 'HMXB_CAND'
    return data


import torch
import pytorch_lightning as L
from torch.utils.data import Dataset
from torchvision import datasets
import torch.optim as optim
import torch.nn as nn
import numpy as np

from typing import Union, Tuple


class DeepAutoencoder(nn.Module):
    def __init__(self, shape_0: int) -> None:
        input_dim = shape_0
        hidden_dim1 = 20
        hidden_dim2 = 10
        latent_dim = 2
        super(DeepAutoencoder, self).__init__()
        self.encoder = nn.Sequential(
            nn.Linear(input_dim, hidden_dim1),
            nn.ReLU(),
            nn.Linear(hidden_dim1, hidden_dim2),
            nn.ReLU(),
            nn.Linear(hidden_dim2, latent_dim)
        )
        self.decoder = nn.Sequential(
            nn.Linear(latent_dim, hidden_dim2),
            nn.ReLU(),
            nn.Linear(hidden_dim2, hidden_dim1),
            nn.ReLU(),
            nn.Linear(hidden_dim1, input_dim),
            nn.Sigmoid()
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.encoder(x)
        x = self.decoder(x)
        return x


class MyEncoder(nn.Module):

    def __init__(self, shapes: list, relu: bool = True, dropout: Union[float, None] = None) -> None:
        super().__init__()
        list_of_layers = []
        for i in range(len(shapes) - 2):
            list_of_layers.append(nn.Linear(shapes[i], shapes[i + 1]))
            if relu:
                list_of_layers.append(nn.ReLU())
            if dropout is not None:
                list_of_layers.append(nn.Dropout(p=dropout))
        list_of_layers.append(nn.Linear(shapes[-2], shapes[-1]))

        self.layers = nn.Sequential(*list_of_layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.layers(x)


class MyDecoder(nn.Module):

    def __init__(self, shapes: list, relu: bool = True, dropout: Union[float, None] = None) -> None:
        super().__init__()
        list_of_layers = []
        for i in range(len(shapes) - 2):
            list_of_layers.append(nn.Linear(shapes[-i - 1], shapes[-i - 2]))
            if relu:
                list_of_layers.append(nn.ReLU())
            if dropout is not None:
                list_of_layers.append(nn.Dropout(p=dropout))
        list_of_layers.append(nn.Linear(shapes[1], shapes[0]))
        self.layers = nn.Sequential(*list_of_layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.layers(x)


class MyAutoencoder(L.LightningModule):

    def __init__(self, shapes: list, relu: bool = True, dropout_en: Union[float, None] = None,
                 dropout_de: Union[float, None] = None) -> None:
        super().__init__()
        self.encoder = MyEncoder(shapes, relu, dropout_en)
        self.decoder = MyDecoder(shapes, relu, dropout_de)

        self.val_outputs: list = []
        self.train_outputs: list = []
        self.train_loss: list = []
        self.val_loss: list = []

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.encoder(x)
        x = self.decoder(x)
        return x

    def calc_loss(self, x: torch.Tensor, x_hat: torch.Tensor) -> torch.Tensor:
        loss = nn.functional.mse_loss(x_hat, x)
        return loss

    def training_step(self, batch: Tuple[torch.Tensor, torch.Tensor], batch_idx: int) -> dict[str, torch.Tensor]:
        x, y = batch
        x = x.view(x.size(0), -1)
        y = y.view(y.size(0), -1)
        x_hat = self.forward(x)
        loss = self.calc_loss(x_hat, y)
        self.log("train_loss", loss, on_step=False, on_epoch=True, prog_bar=True, logger=False)
        self.train_outputs.append({'loss': loss})
        return {'loss': loss}

    def validation_step(self, batch: Tuple[torch.Tensor, torch.Tensor], batch_idx: int) -> dict[str, torch.Tensor]:
        with torch.no_grad():
            x, y = batch
            x = x.view(x.size(0), -1)
            x_hat = self.forward(x)
            loss = self.calc_loss(x_hat, y)
        self.log("val_loss", loss, prog_bar=True)
        self.val_outputs.append({'loss': loss})
        return {'loss': loss}

    def configure_optimizers_(self) -> Tuple[list[torch.optim.Optimizer], list[dict[str, Any]]]:
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-3, weight_decay=5e-4)

        lr_scheduler = {
            'scheduler': torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='min', factor=0.5, patience=10, min_lr=1e-6, verbose=True),
            'monitor': 'val_loss',
        }

        return [optimizer], [lr_scheduler]

    def configure_optimizers(self) -> Tuple[list[torch.optim.Optimizer], list[dict[str, Any]]]:
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-3, weight_decay=5e-4)

        # Define a custom lambda function for the sawtooth learning rate
        def lr_lambda(epoch: int) -> float:
            # Define the sawtooth shape for the learning rate
            cycle = 10  # Number of iterations in a cycle
            return 0.5 * (1 - abs(epoch % cycle - cycle // 2) / (cycle // 2))

        lr_scheduler = {
            'scheduler': torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda),
            'interval': 'epoch',
        }

        return [optimizer], [lr_scheduler]

    def on_validation_epoch_end(self) -> None:
        # здесь вы можете использовать self.train_outputs для доступа к выводам обучения
        avg_loss = torch.stack([x['loss'] for x in self.val_outputs]).mean()
        print(f"| Val Loss: {avg_loss:.3f}")
        self.log('val_loss', avg_loss, prog_bar=True, on_epoch=True, on_step=False)
        self.val_outputs = []
        self.val_loss.append(avg_loss)

    def on_train_epoch_end(self) -> None:
        # здесь вы можете использовать self.train_outputs для доступа к выводам обучения
        avg_loss = torch.stack([x['loss'] for x in self.train_outputs]).mean()
        print(f"| Train Loss: {avg_loss:.3f}")
        self.log('train_loss', avg_loss, prog_bar=True, on_epoch=True, on_step=False)
        self.train_outputs = []
        self.train_loss.append(avg_loss)


class MyDataset(Dataset):
    def __init__(self, dataset: np.ndarray) -> None:
        self.dataset = dataset
        self.dataset = torch.tensor(self.dataset, dtype=torch.float)

    def __len__(self) -> int:
        return len(self.dataset)

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor, torch.Tensor]:
        return self.dataset[idx], self.dataset[idx]


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from ..utils import get_attrs_columns, merge_save_attrs, correlate


def _get_near_pairs(df, max_distance=5, percentile=0.99, plot=False):
    first = df[get_attrs_columns(df, ['id', 'ra', 'dec', 'poserr'])].copy()
    second = df[get_attrs_columns(df, ['id', 'ra', 'dec', 'poserr'])].copy()
    second.attrs['name'] = 'second'
    first.attrs['name'] = 'first'
    cor = correlate(first, second, max_distance, add_distance=True)
    cor.query(f'{df.attrs["id"]+"_first"} != {df.attrs["id"]+"_second"}', inplace=True)
    if len(cor)==0: 
        if plot:
            print(f'No pairs found in {max_distance} arcsec')
        return []
    cor['percentile_first'] = (-2 * (cor[df.attrs['poserr']+'_first']**2) * np.log(1-percentile))**0.5
    cor['percentile_second'] = (-2 * (cor[df.attrs['poserr']+'_second']**2) * np.log(1-percentile))**0.5
    if plot: cor['distance'].hist(bins=max(100, int(cor['distance'].max()/10)))
    cor.query('percentile_first+percentile_second>distance', inplace=True)
    
    
    conn = cor[[df.attrs['id']+'_first', df.attrs['id']+'_second']].values.tolist()
    conn = list(map(lambda x: tuple(x), conn))
    conn = list(set(conn))
    return conn

def _connected_components(neighbors):
    seen = set()
    def component(node):
        nodes = set([node])
        while nodes:
            node = nodes.pop()
            seen.add(node)
            nodes |= neighbors[node] - seen
            yield node
    for node in neighbors:
        if node not in seen:
            yield component(node)

def _get_component_to_id(pairs):
    graph = dict()

    for v1, v2 in pairs:
        if v1 in graph:
            graph[v1].add(v2)
        else:
            graph[v1] = set([v2])
        if v2 in graph:
            graph[v2].add(v1)
        else:
            graph[v2] = set([v1])

    ans = dict()
    for num, components in enumerate(_connected_components(graph)):
        for comp in components:
            ans[comp] = num

    clasters = pd.DataFrame(ans.items(), columns=['id', 'claster'])
    return clasters

def _prepare_agg_plan(df, provided_plan=dict()):
    #provided_plan = {'min': ['u_magErr', 'g_magErr', 'r2_magErr', 'ha_magErr', 'r_magErr', 'i_magErr']}
    plan = {'first': get_attrs_columns(df, ['id', 'ra', 'dec', 'poserr'])}
    if 'mags' in df.attrs:
        plan['min'] = df.attrs['mags']

    agg = {}
    for key, val in plan.items():
        for i in val:
            agg[i] = key
    for key, val in provided_plan.items():
        for i in val:
            agg[i] = key
    for col in df.columns:
        if col not in agg:
            agg[col] = 'median'
    return agg

def merge_duplicates_nearest(df: pd.DataFrame, max_distance=5, percentile=0.99,
                             agg_functions=dict(), verbose=False) -> pd.DataFrame:
    """
    {'min': ['u_magErr', 'g_magErr', 'r2_magErr', 'ha_magErr', 'r_magErr', 'i_magErr']}
    """
    if verbose: print('Size before transformation:', len(df))
    pairs = _get_near_pairs(df, max_distance=max_distance, percentile=percentile, plot=False)
    clasters = _get_component_to_id(pairs)
    agg_task = _prepare_agg_plan(df, agg_functions)

    problems = df[df[df.attrs['id']].isin(clasters['id'])].copy()
    problems = merge_save_attrs(problems, clasters.rename(columns={'id': problems.attrs['id']}), on=problems.attrs['id'])
    problems = problems.sort_values(problems.attrs['poserr']).groupby('claster').agg(agg_task)
    not_problem = df[~df[df.attrs['id']].isin(clasters['id'])].copy()

    res = pd.concat([not_problem, problems.reset_index(drop=True)]).reset_index(drop=True)
    if verbose: print('Size after transformation:', len(res))
    return res

def merge_duplicates_same_id(df: pd.DataFrame, agg_functions=dict(), verbose=False) -> pd.DataFrame:
    if verbose: print('Size before transformation:', len(df))
    
    agg_task = _prepare_agg_plan(df, agg_functions)
    if df.attrs['id'] in agg_task: del agg_task[df.attrs['id']]

    res = df.groupby(df.attrs['id']).agg(agg_task).reset_index()
    if verbose: print('Size after transformation:', len(res))
    return res

def merge_duplicates_same_coords(df: pd.DataFrame, agg_functions=dict(), verbose=False) -> pd.DataFrame:
    if verbose: print('Size before transformation:', len(df))

    agg_task = _prepare_agg_plan(df, agg_functions)
    agg_task[df.attrs['id']] = 'first'
    if df.attrs['ra'] in agg_task: del agg_task[df.attrs['ra']]
    if df.attrs['dec'] in agg_task: del agg_task[df.attrs['dec']]

    res = df.groupby([df.attrs['ra'], df.attrs['dec']]).agg(agg_task).reset_index()
    if verbose: print('Size after transformation:', len(res))
    return res

def plot_duplicates_problem(df, max_distance=5, ax=None, title=''):
    if ax is None:
        fig, ax = plt.subplots()
    cor = correlate(df, df, max_distance, add_distance=True).query('distance!=0')
    cor['distance'].plot(kind='hist', bins=max(100, int(cor['distance'].max()/10)), ax=ax, title=title)